"""
Model classes for AppDynamics REST API

.. moduleauthor:: Kyle Furlong <kyle.furlong@appdynamics.com>
"""

from . import JsonObject, JsonList


class Backend(JsonObject):

    FIELDS = {'id': '', 'name': '', 'exit_point_type': 'exitPointType', 'properties': ''}
    EXIT_POINT_TYPES = ('HTTP', 'CACHE', 'DB', 'WEB_SERVICE', 'MODULE')

    def __init(self, backend_id=0, name='', exit_point_type='', properties=''):
        self._exit_point_type = None
        self.id, self.name, self.exit_point_type, self.properties, = \
            backend_id, name, exit_point_type, properties

    @property
    def exit_point_type(self):
        """
        :rtype: str
        """
        return self._exit_point_type

    @exit_point_type.setter
    def exit_point_type(self, exit_point_type):
        self._list_setter('_exit_point_type', exit_point_type, Backend.EXIT_POINT_TYPES)


class Backends(JsonList):

    def __init__(self, initial_list=None):
        super(Backends, self).__init__(Backend, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Backend
        """
        return self.data[i]

    def by_exit_point_type(self, exit_point_type):
        """
        Searches for backends of a particular type (which should be one of Backend.EXIT_POINT_TYPES).
        For example, to find all the HTTP backends:

        >>> from appd.request import AppDynamicsClient
        >>> client = AppDynamicsClient(...)
        >>> all_backends = client.get_backends()
        >>> http_backends = all_backends.by_exit_point_type('HTTP')

        :returns: a Tiers object containing any tiers matching the criteria
        :rtype: Tiers
        """
        return Backends([x for x in self.data if x.exit_point_type == exit_point_type])



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class Node(JsonObject):

    FIELDS = {'id': '', 'name': '', 'type': '', 'machine_id': 'machineId', 'machine_name': 'machineName',
              'tier_id': 'tierId', 'tier_name': 'tierName', 'unique_id': 'nodeUniqueLocalId',
              'os_type': 'machineOSType', 'has_app_agent': 'appAgentPresent', 'app_agent_version': 'appAgentVersion',
              'has_machine_agent': 'machineAgentPresent', 'machine_agent_version': 'machineAgentVersion'}

    def __init__(self, node_id=0, name='', node_type='', machine_id=0, machine_name='', os_type='',
                 unique_local_id='', tier_id=0, tier_name='', has_app_agent=False, app_agent_version='',
                 has_machine_agent=False, machine_agent_version=''):
        (self.id, self.name, self.type, self.machine_id, self.machine_name, self.os_type, self.unique_local_id,
         self.tier_id, self.tier_name, self.has_app_agent, self.app_agent_version, self.has_machine_agent,
         self.machine_agent_version) = (node_id, name, node_type, machine_id, machine_name, os_type, unique_local_id,
                                        tier_id, tier_name, has_app_agent, app_agent_version,
                                        has_machine_agent, machine_agent_version)


class Nodes(JsonList):

    def __init__(self, initial_list=None):
        super(Nodes, self).__init__(Node, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Node
        """
        return self.data[i]

    def by_machine_name(self, name):
        """
        Filters a Nodes collection to return only the nodes matching the given hostname.
        :param str name: Hostname to match against.
        :returns: a Nodes collection filtered by hostname.
        :rtype: Nodes
        """
        return Nodes([x for x in self.data if x.machineName == name])

    def by_machine_id(self, machine_id):
        """
        Filters a Nodes collection to return only the nodes matching the given machine instance ID.
        :param int machine_id: Machine ID to match against.
        :returns: a Nodes collection filtered by machine ID.
        :rtype: Nodes
        """
        return Nodes([x for x in self.data if x.machineId == machine_id])

    def by_tier_name(self, name):
        """
        Filters a Nodes collection to return only the nodes belonging to the given tier.
        :param str name: Tier name to match against.
        :returns: a Nodes collection filtered by tier.
        :rtype: Nodes
        """
        return Nodes([x for x in self.data if x.tier_name == name])

    def by_tier_id(self, tier_id):
        """
        Filters a Nodes collection to return only the nodes belonging to the given tier ID.
        :param int tier_id: Tier ID to match against.
        :returns: a Nodes collection filtered by tier.
        :rtype: Nodes
        """
        return Nodes([x for x in self.data if x.tier_id == tier_id])



"""
Model classes for AppDynamics REST API
"""

from . import JsonObject, JsonList


class LicenseModule(JsonObject):

    FIELDS = {'name': ''}

    def __init__(self, name=None):
        self.name = name


class LicenseModuleList(JsonList):
    """
    Represents a collection of :class:Account objects. Extends :class:UserList, so it supports the
    standard array index and :keyword:`for` semantics.
    """
    def __init__(self, initial_list=None):
        super(LicenseModuleList, self).__init__(LicenseModule, initial_list)

    def __getitem__(self, i):
        """
        :rtype: LicenseModule
        """
        return self.data[i]

    def by_name(self, name):
        """
        Finds an account by name.

        :returns: First account with the correct name
        :rtype: LicenseModule
        """
        found = [x for x in self.data if x.name == name]
        try:
            return found[0]
        except IndexError:
            raise KeyError(name)

    def __contains__(self, item):
        found = [x for x in self.data if x.name == item]
        return True if found else False


class LicenseModules(JsonObject):

    FIELDS = {}

    def __init__(self):
        self.modules = LicenseModuleList()

    @classmethod
    def from_json(cls, json_dict):
        obj = super(LicenseModules, cls).from_json(json_dict)
        obj.modules = LicenseModuleList.from_json(json_dict['modules'])
        return obj




"""
Model classes for AppDynamics REST API
"""

from . import JsonObject, JsonList
from appd.time import from_ts

class HourlyLicenseUsage(JsonObject):

    FIELDS = {
        'id': '',
        'account_id': 'accountId',
        'max_units_used': 'maxUnitsUsed',
        'min_units_used': 'minUnitsUsed',
        'avg_units_used': 'avgUnitsUsed',
        'total_units_used': 'totalUnitsUsed',
        'sample_count': 'sampleCount',
        'avg_units_allowed': 'avgUnitsAllowed',
        'avg_units_provisioned': 'avgUnitsProvisioned',
        'license_module': 'agentType',
        'created_on_ms': 'createdOn',
    }

    def __init__(self, id=0, account_id=0, max_units_used=0, min_units_used=0, avg_units_used=0, total_units_used=0,
                 sample_count=0, avg_units_allowed=0, avg_units_provisioned=None, license_module=None,
                 created_on_ms=0):
        (self.id, self.account_id, self.max_units_used, self.min_units_used, self.avg_units_used, self.total_units_used,
         self.sample_count, self.avg_units_allowed, self.avg_units_provisioned, self.license_module,
         self.created_on_ms) = (id, account_id, max_units_used, min_units_used, avg_units_used, total_units_used,
                                sample_count, avg_units_allowed, avg_units_provisioned, license_module,
                                created_on_ms)

    @property
    def created_on(self):
        """
        :rtype: datetime.datetime
        """
        return from_ts(self.created_on_ms)


class HourlyLicenseUsageList(JsonList):

    def __init__(self, initial_list=None):
        super(HourlyLicenseUsageList, self).__init__(HourlyLicenseUsage, initial_list)

    def __getitem__(self, i):
        """
        :rtype: HourlyLicenseUsage
        """
        return self.data[i]

    def by_account_id(self, account_id):
        """
        :rtype: HourlyLicenseUsageList
        """
        return HourlyLicenseUsageList([x for x in self.data if x.account_id == account_id])

    def by_license_module(self, license_module):
        """
        :rtype: HourlyLicenseUsageList
        """
        return HourlyLicenseUsageList([x for x in self.data if x.license_module == license_module])


class HourlyLicenseUsages(JsonObject):

    FIELDS = {}

    def __init__(self):
        self.usages = HourlyLicenseUsageList()

    @classmethod
    def from_json(cls, json_dict):
        obj = super(HourlyLicenseUsages, cls).from_json(json_dict)
        obj.usages = HourlyLicenseUsageList.from_json(json_dict['usages'])
        return obj




#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for AppDynamics REST API
"""

import unittest
import appd
import creds.localhost as creds

__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


class ApplicationApiTest(unittest.TestCase):

    def setUp(self):
        self.c = appd.request.AppDynamicsClient(creds.url, creds.user, creds.password, creds.account)

    def assertIn(self, test_value, expected_set):
        """
        Included for compatibility with Python 2.6
        """
        msg = "%s did not occur in %s" % (test_value, expected_set)
        self.assert_(test_value in expected_set, msg)

    def assertNotIn(self, test_value, expected_set):
        """
        Included for compatibility with Python 2.6
        """
        msg = "%s was found in %s" % (test_value, expected_set)
        self.assert_(test_value not in expected_set, msg)


if __name__ == '__main__':
    from test_v1_applications import *
    from test_v2_accounts import *
    from test_v2_license_modules import *
    suite = unittest.TestLoader()\
        .loadTestsFromTestCase(V1_ApplicationTest)\
        .loadTestsFromTestCase(V2_AccountTest)\
        .loadTestsFromTestCase(V2_LicenseModuleTest)

    unittest.TextTestRunner(verbosity=2).run(suite)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for AppDynamics REST API
"""

from test import ApplicationApiTest


class V2_AccountTest(ApplicationApiTest):

    def test_my_account(self):
        acct = self.c.get_my_account()
        self.assertEqual(acct.id, '2')
        self.assertEqual(acct.name, 'customer1')

    def test_get_account_by_id(self):
        acct = self.c.get_account(2)
        self.assertEqual(acct.id, '2')
        self.assertEqual(acct.name, 'customer1')



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList
from appd.time import from_ts

class MetricValue(JsonObject):

    FIELDS = {
        'occurrences': '',
        'current': '',
        'min': '',
        'max': '',
        'start_time_ms': 'startTimeInMillis',
        'count': '',
        'sum': '',
        'value': '',
        'std_dev': 'standardDeviation'
    }

    def __init__(self, occurrences=0, current=0, min_value=0, max_value=0, start_time_ms=0, count=0, sum=0, value=0, std_dev=0):
        (self.occurrences, self.current, self.min, self.max, self.start_time_ms, self.count, self.sum, self.value, self.std_dev) = (occurrences, current, min_value, max_value, start_time_ms, count, sum, value, std_dev)

    @property
    def start_time(self):
        """
        Gets the timestamp of the metric data, converting it from an AppDynamics timestamp to standard
        Python :class:`datetime`.

        :return: Time the violation was resolved
        :rtype: :class:`datetime.datetime`
        """
        return from_ts(self.start_time_ms)


class MetricValues(JsonList):

    def __init__(self, initial_list=None):
        super(MetricValues, self).__init__(MetricValue, initial_list)

    def __getitem__(self, i):
        """
        :rtype: MetricValue
        """
        return self.data[i]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to print a list of events from the specified controller and app.
"""

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient

__author__ = 'Kyle Furlong'
__copyright__ = 'Copyright (c) 2013-2017 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

resp = c.get_audit_history('2017-04-27T08:00:00.000-0800', '2017-04-27T20:00:00.000-0800')

for audit in resp.by_user_name("user1"):
    print(audit)
for audit in resp.by_action("LOGIN"):
    print(audit)
for audit in resp.by_account_name("customer1"):
    print(audit)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This script generates a report of response times, calls, and errors for all the BT's registered
on a controller. The output is generated as XML and can be sent to a file with this syntax::

    python BusinessTransactionsMetrics.py > metric_output.xml

By default this will connect to a controller on localhost:8090. The script accepts command-line
arguments to change the connection parameters:

--controller=<url>
--account=<account>
--username=<name>
--password=<password>

Example::
     python BTMetrics.py --controller=http://10.1.2.3:8090 --account=customer1 --username=demo --password=abc123
"""

from __future__ import print_function

from collections import defaultdict
from datetime import datetime
from time import mktime
from lxml.builder import ElementMaker
from lxml import etree
import tzlocal

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


# The report will generate data for the 24-hour period before midnight of the current day. To change the
# reporting period, adjust these variables.

time_in_mins = 1440
end_time = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
end_epoch = int(mktime(end_time.timetuple())) * 1000


# Helper functions

def now_rfc3339():
    return datetime.now(tzlocal.get_localzone()).isoformat('T')


# Parse command line arguments and create AD client:

args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)


# Get the list of configured apps, and get backend metrics for each one:

METRIC_MAP = {'Average Block Time (ms)': 'abt',
              'Average CPU Used (ms)': 'cpu',
              'Average Request Size': 'req_size',
              'Average Response Time (ms)': 'art',
              'Average Wait Time (ms)': 'wait_time',
              'Calls per Minute': 'cpm',
              'End User Average Response Time (ms)': 'eum_art',
              'End User Network Average Response Time (ms)': 'eum_net',
              'End User Page Render Average Response Time (ms)': 'eum_render',
              'Errors per Minute': 'epm',
              'Normal Average Response Time (ms)': 'norm_art',
              'Number of Slow Calls': 'slow',
              'Number of Very Slow Calls': 'veryslow',
              'Stall Count': 'stalls'}


empty_row = dict([(x, 0) for x in list(METRIC_MAP.values())])
rows = defaultdict(dict)

for app in c.get_applications():

    bt_list = c.get_bt_list(app.id)

    for md in c.get_metrics('Business Transaction Performance|Business Transactions|*|*|*',
                            app.id, time_range_type='BEFORE_TIME', end_time=end_epoch,
                            duration_in_mins=time_in_mins, rollup=True):

        # Get the last 3 components of the metric path. This should be 'tier_name|bt_name|metric_name'.
        tier_name, bt_name, metric_name = md.path.split('|')[-3:]
        tier_bts = bt_list.by_tier_and_name(bt_name, tier_name)
        if tier_bts:
            bt = tier_bts[0]
            if len(md.values) > 0 and metric_name in METRIC_MAP:
                key = (tier_name, bt_name)
                rows.setdefault(key, empty_row.copy()).update({'app_id': app.id,
                                                               'app_name': app.name,
                                                               'bt_id': bt.id,
                                                               'bt_name': bt.name,
                                                               'tier_name': bt.tier_name,
                                                               'type': bt.type,
                                                               METRIC_MAP[metric_name]: md.values[0].value})


# Generate the report.

XSI = 'http://www.w3.org/2001/XMLSchema-instance'
E = ElementMaker(nsmap={'xsi': XSI})

root = E.BusinessTransactions(Controller=c.base_url, GenerationTime=now_rfc3339())
root.set('{%s}noNamespaceSchemaLocation' % XSI, 'bt_metrics.xsd')

for k, v in sorted(rows.items()):
    v['calls'] = v['cpm'] * time_in_mins
    v['errors'] = v['epm'] * time_in_mins
    v['error_pct'] = round(float(v['errors']) / float(v['calls']) * 100.0, 1) if v['calls'] > 0 else 0

    root.append(E.BusinessTransaction(
        E.ApplicationName(v['app_name']),
        E.BusinessTransactionName(v['bt_name']),
        E.TierName(v['tier_name']),
        E.AverageResponseTime(str(v['art'])),
        E.CallsPerMinute(str(v['cpm'])),
        E.TotalCalls(str(v['calls'])),
        E.TotalErrors(str(v['errors'])),
        E.ErrorsPerMinute(str(v['epm'])),
        E.ErrorPercentage(str(v['error_pct'])),
        E.SlowCalls(str(v['slow'])),
        E.VerySlowCalls(str(v['veryslow'])),
        E.Stalls(str(v['stalls'])),
    ))


# Print the report to stdout.

print(etree.ProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"'))
print(etree.tostring(root, pretty_print=True, encoding='UTF-8'))



"""
This module contains the main classes for handling requests to the AppDynamics REST API.
"""

from __future__ import print_function

import sys
import json
import requests

from datetime import datetime

from urllib import parse

from appd.model.account import *
from appd.model.application import *
from appd.model.config_variable import *
from appd.model.license_module import *
from appd.model.hourly_license_usage import *
from appd.model.license_usage import *
from appd.model.tier import *
from appd.model.backend import *
from appd.model.metric_treenode import *
from appd.model.business_transaction import *
from appd.model.policy_violation import *
from appd.model.snapshot import *
from appd.model.metric_data import *
from appd.model.node import *
from appd.model.set_controller_url import *
from appd.model.event import *
from appd.model.action_suppressions import *
from appd.model.audit_history import *


class AppDynamicsClient(object):
    """
    Main class that wraps around the REST API to make it easier to send requests
    and parse responses.
    """

    TIME_RANGE_TYPES = ('BEFORE_NOW', 'BEFORE_TIME', 'AFTER_TIME', 'BETWEEN_TIMES')

    DEEP_DIVE_POLICIES = ('SLA_FAILURE', 'TIME_SAMPLING', 'ERROR_SAMPLING', 'OCCURRENCE_SAMPLING',
                          'ON_DEMAND', 'HOTSPOT', 'HOTSPOT_LEARN', 'APPLICATION_STARTUP', 'DIAGNOSTIC_SESSION',
                          'SLOW_DIAGNOSTIC_SESSION', 'ERROR_DIAGNOSTIC_SESSION', 'POLICY_FAILURE_DIAGNOSTIC_SESSION',
                          'INFLIGHT_SLOW_SESSION')

    USER_EXPERIENCES = ('NORMAL', 'SLOW', 'VERY_SLOW', 'STALL', 'BUSINESS_ERROR')

    COLLECTOR_TYPES = ('ERROR_IDS', 'STACK_TRACES', 'ERROR_DETAIL', 'HTTP_PARAMETER', 'BUSINESS_DATA',
                       'COOKIE', 'HTTP_HEADER', 'SESSION_KEY', 'RESPONSE_HEADER', 'LOG_MESSAGE',
                       'TRANSACTION_PROPERTY', 'TRANSACTION_EVENT', 'DOTNET_PROPS', 'DOTNET_SET')

    SNAPSHOT_REQUEST_PARAMS = ('guids', 'archived', 'deep-dive-policy', 'application-component-ids',
                               'application-component-node-ids', 'business-transaction-ids', 'user-experience',
                               'first-in-chain', 'need-props', 'need-exit-calls', 'execution-time-in-millis',
                               'session-id', 'user-principal-id', 'error-ids', 'error-occurred',
                               'bad-request', 'diagnostic-snapshot', 'diagnostic-session-guid',
                               'starting-request-id', 'ending-request-id',
                               'data-collector-name', 'data-collector-type', 'data-collector-value')

    SNAPSHOT_REQUEST_LISTS = ('business-transaction-ids', 'user-experience', 'error-ids', 'guids', 'deep-dive-policy',
                              'application-component-ids', 'application-component-node-ids', 'diagnostic-session-guid')

    CPM = CALLS_PER_MINUTE = 'Calls per Minute'
    ART = AVERAGE_RESPONSE_TIME = 'Average Response Time (ms)'
    NORMAL_ART = 'Normal Average Response Time (ms)'
    SLOW_CALLS = 'Number of Slow Calls'
    VERY_SLOW_CALLS = 'Number of Very Slow Calls'
    EPM = ERRORS_PER_MINUTE = 'Errors per Minute'
    EXCEPTIONS_PER_MINUTE = 'Exceptions per Minute'
    STALLS = 'Stall Count'

    def __init__(self, base_url='http://localhost:8090', username='user1', password='welcome',
                 account='customer1', debug=False, verify=True):
        """
        Creates a new instance of the client.

        :param base_url: URL of your controller.
        :type base_url: str.
        :param username: User name to authenticate to the controller with.
        :type username: str.
        :param password: Password for authentication to the controller.
        :type password: str.
        :param account: Account name for multi-tenant controllers. For single-tenant controllers, use
                        the default value of "customer1".
        :param debug: Set to :const:`True` to print extra debugging information to :const:`sys.stdout`.
        :type debug: bool.
        """

        self._username, self._password, self._account, self._app_id, self._session = '', '', '', None, None
        self._base_url = ''
        (self.base_url, self.username, self.password, self.account, self.debug, self.verify) = (base_url, username, password,
                                                                                   account, debug, verify)

    @property
    def base_url(self):
        return self._base_url

    @base_url.setter
    def base_url(self, new_url):
        self._base_url = new_url
        if '://' not in self._base_url:
            self._base_url = 'http://' + self._base_url
        while self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def __make_auth(self):
        self._auth = (self._username + '@' + self._account, self._password)

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, new_username):
        self._username = new_username
        self.__make_auth()

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, new_password):
        self._password = new_password
        self.__make_auth()

    @property
    def account(self):
        return self._account

    @account.setter
    def account(self, new_account):
        self._account = new_account
        self.__make_auth()

    @property
    def app_id(self):
        return self._app_id

    @app_id.setter
    def app_id(self, new_app_id):
        self._app_id = new_app_id

    def _get_session(self):
        if not self._session:
            from requests.sessions import Session
            self._session = Session()
            self._session.verify = self.verify
        return self._session

    def _request(self, **request_args):
        s = self._get_session()
        req = requests.Request(**request_args)

        prepped = s.prepare_request(req)

        # Merge environment settings into session
        settings = s.merge_environment_settings(prepped.url, {}, None, None, None)

        return s.send(prepped, **settings)

    def upload(self, path, data, filename='default'):
        """
        Workaround a bug between requests file upload boundary format
        and the expected format at the server
        """

        import pycurl
        import base64
        try:
            from io import BytesIO
        except ImportError:
            from StringIO import StringIO as BytesIO

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        c = pycurl.Curl()
        b = BytesIO()

        c.setopt(c.SSL_VERIFYPEER, int(self.verify))
        c.setopt(c.SSL_VERIFYHOST, int(self.verify))

        c.setopt(c.WRITEDATA, b)
        c.setopt(c.URL, url)

        auth = base64.encodestring('%s:%s' % self._auth).replace('\n', '')

        c.setopt(c.HTTPHEADER, [
            'Authorization: Basic {0}'.format(auth)
        ])
        c.setopt(c.HTTPPOST, [
            ('fileupload', (
                c.FORM_BUFFER, filename,
                c.FORM_BUFFERPTR, data,
            )),
        ])

        c.perform()
        c.close()

        return b.getvalue()

    def request(self, path, params=None, method='GET', use_json=True, query=True, headers=None):
        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        params = params or {}
        if use_json and method == 'GET':
            params['output'] = 'JSON'
        for k in list(params.keys()):
            if params[k] is None:
                del params[k]

        if self.debug:
            print('Retrieving ' + url, self._auth, params)

        if method == 'GET' or query:
            r = self._request(method=method, url=url, auth=self._auth, params=params, headers=headers)
        else:
            if not headers:
                headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
            r = self._request(method=method, url=url, auth=self._auth, data=json.dumps(params), headers=headers)

        if r.status_code != requests.codes.ok:
            print(url, file=sys.stderr)
            r.raise_for_status()

        return r.json() if use_json else r.text

    def _app_path(self, app_id, path=None):
        app_id = app_id if isinstance(app_id, int) or isinstance(app_id, str) else self._app_id
        if not app_id:
            raise ValueError('application id is required')
        path = '/controller/rest/applications/%s' % parse.quote(str(app_id)) + (path or '')
        return path

    def get_metric_tree(self, app_id=None, metric_path=None, recurse=False):
        """
        Retrieves a list of available metrics.

        :param int app_id: Application ID to retrieve metrics for. If :const:`None`, the value stored in the
          `app_id` property will be used.
        :param str metric_path: Point in the metric tree at which you want to retrieve the available metrics.
          If :const:`None`, start at the root.
        :param bool recurse: If :const:`True`, retrieve the entire tree from :data:`metric_path` on down.
        :returns: An object containing the metrics under this point in the tree.
        :rtype: appd.model.MetricTreeNodes
        """
        parent = None
        if metric_path is not None:
            parent = MetricTreeNode(parent=None, node_name=metric_path, node_type='folder')
        return self._get_metric_tree(app_id, parent=parent, recurse=recurse)

    def _get_metric_tree(self, app_id=None, parent=None, recurse=False):
        params = {}
        if parent is not None:
            params['metric-path'] = parent.path
        path = self._app_path(app_id, '/metrics')
        nodes = MetricTreeNodes.from_json(self.request(path, params), parent)
        if recurse:
            for node in nodes:
                if node.type == 'folder':
                    self._get_metric_tree(app_id, parent=node, recurse=True)
        return nodes

    # Top-level requests

    def _top_request(self, cls, path, params=None, method='GET', query=True):
        return cls.from_json(self.request('/controller/rest' + path, params, method, query=query))

    def set_controller_url(self, controllerURL):
        """
        Sets the base url for all email notification links

        :param str controller_url: Base URL to use for email notification links
        :returns: Success message
        """
        param = {'controllerURL': controllerURL}
        return self._top_request(SetControllerUrlResponse,
                                 '/accounts/%s/update-controller-url' % self.account, param, 'POST', query=False)

    def get_audit_history(self, start_time, end_time, time_zone_id=None):
        """
        Gathers audit history for up to a 24h time period

        :param str start_time: Start time in the format: "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
        :param str end_time: End time in the format: "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
        :param str time_zone_id: Optional, time zone in tz format

        :return: A :class:`AuditHistory <appd.model.AuditHistory>` object, representing a collection of audits.
        :rtype: appd.model.AuditHistory
        """

        params = {
            'startTime': start_time,
            'endTime': end_time
        }

        if (time_zone_id):
            params.update({'timeZoneId': time_zone_id})

        return AuditHistory.from_json(self.request('/controller/ControllerAuditHistory',
                                                   params, 'GET', query=True, use_json=True))

    def create_user(self,
                    user_name,
                    user_display_name,
                    user_email,
                    user_password,
                    user_roles=None):
        """
        Creates a user

        :param str user_name: User's name (required)
        :param str user_display_name: How the user's name will show up in the UI (required)
        :param str user_email: User's email (required)
        :param str user_roles: Comma-separated list of roles
            https://docs.appdynamics.com/display/PRO42/Roles+and+Permissions#RolesandPermissions-PredefinedRoles
        :param str user_password: Required on create, not on update

        :returns: Nothing on success, exception with HTTP error code on failure
        """
        params = {'user-name': user_name,
                  'user-display-name': user_display_name,
                  'user-email': user_email}

        if (user_roles):
            params.update({'user-roles': user_roles})

        if (user_password):
            params.update({'user-password': user_password})

        return self.request('/controller/rest/users', params, 'POST', query=True, use_json=False)

    def export_health_rules(self, application_id, name=None):
        """
        Exports all health rules from the given app, in XML format

        :param int application_id: Application ID
        :param string name: Name of a particular health rule to export,
            if omitted, all health rules will be exported
        :returns: JSON string
        """
        params = {}
        if (name):
            params.update({'name': name})

        return self.request('/controller/healthrules/{0}'.format(application_id),
                            params, 'GET', query=True, use_json=False)

    def import_health_rules(self, application_id, xml, overwrite=False):
        """
        Imports all health rules into the given app, from XML format

        :param int application_id: Application ID
        :param string xml: Output of export_health_rules

        :returns: Plain text string, containing success or failure messages
        """

        params = {}
        if (overwrite):
            params.update({'overwrite': overwrite})

        path = '/controller/healthrules/{0}'.format(application_id)

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('healthrules.xml', xml)}
        r = self._request(method='POST', url=url, auth=self._auth, params=params, files=files)
        return r.text

    def export_entry_point_type(self, application_id, entry_point_type, rule_type, tier=None, rule_name=None):
        """
        Exports one entry point type from the given app, in XML format

        :param int application_id: Application ID
        :param string entry_point_type: type of entry point, see AppDynamics docs
        :param string rule_type: one of auto, custom, or exclude
        :param string tier: name of the tier to export from, optional

        :returns: XML string
        """
        url = '/controller/transactiondetection/{0}/'.format(application_id)

        if (tier):
            url = url + tier + '/'

        url = url + rule_type + '/'
        url = url + entry_point_type

        if (rule_name):
            url = url + '/' + rule_name

        return self.request(url, {}, 'GET', query=True, use_json=False)

    def import_entry_point_type(self, application_id, entry_point_type, rule_type, xml, tier=None, rule_name=None):
        """
        Imports all entry points into the given app, from XML format

        :param int application_id: Application ID
        :param string rule_type: one of auto, custom, or exclude
        :param string xml: Output of export_entry_points
        :param string tier: name of the tier to export from, optional

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/transactiondetection/{0}/'.format(application_id)

        if (tier):
            path = path + tier + '/'

        path = path + rule_type + '/'
        path = path + entry_point_type

        if (rule_name):
            path = path + '/' + rule_name

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('entrypoints.xml', xml)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def export_entry_points(self, application_id, rule_type, tier=None):
        """
        Exports all entry points from the given app, in XML format

        :param int application_id: Application ID
        :param string rule_type: one of auto, custom, or exclude
        :param string tier: name of the tier to export from, optional

        :returns: XML string
        """
        url = '/controller/transactiondetection/{0}/'.format(application_id)

        if (tier):
            url = url + tier + '/'

        url = url + rule_type

        return self.request(url, {}, 'GET', query=True, use_json=False)

    def import_entry_points(self, application_id, rule_type, xml, tier=None):
        """
        Imports all entry points into the given app, from XML format

        :param int application_id: Application ID
        :param string rule_type: one of auto, custom, or exclude
        :param string xml: Output of export_entry_points
        :param string tier: name of the tier to export from, optional

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/transactiondetection/{0}/'.format(application_id)

        if (tier):
            path = path + tier + '/'

        path = path + rule_type

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('entrypoints.xml', xml)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def export_analytics_dynamic_service_configs(self, application_id):
        """
        Exports all analytics dynamics service configs from the given app, in XML format
        Only available on controllers 4.3+

        :param int application_id: Application ID

        :returns: JSON string
        """

        return self.request('/controller/analyticsdynamicservice/{0}'.format(application_id),
                            {}, 'GET', query=True, use_json=False)

    def import_analytics_dynamic_service_configs(self, application_id, xml):
        """
        Imports all actions into the given app, from JSON format
        Only available on controllers 4.3+

        :param int application_id: Application ID
        :param string xml: Output of export_analytics_dynamic_service_configs

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/analyticsdynamicservice/{0}'.format(application_id)

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('actions.xml', xml)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def export_custom_dashboard(self, dashboard_id):
        """
        Exports custom dashboard with the given ID, in JSON format

        :param int dashboard_id: custom dashboard ID

        :returns: JSON string
        """

        params = {}
        if (dashboard_id):
            params.update({'dashboardId': dashboard_id})

        return self.request('/controller/CustomDashboardImportExportServlet', params, 'GET', query=True, use_json=False)

    def import_custom_dashboard(self, json):
        """
        Imports custom dashboard, from JSON format

        :param string json: Output of export_custom_dashboard

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/CustomDashboardImportExportServlet'

        return self.upload(path, json)

    def export_policies(self, application_id):
        """
        Exports all policies from the given app, in JSON format

        :param int application_id: Application ID

        :returns: JSON string
        """

        return self.request('/controller/actions/{0}'.format(application_id), {}, 'GET', query=True, use_json=False)

    def import_policies(self, application_id, json):
        """
        Imports all policies into the given app, from JSON format

        :param int application_id: Application ID
        :param string json: Output of export_policies

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/actions/{0}'.format(application_id)

        return self.upload(path, json)

    def export_actions(self, application_id):
        """
        Exports all actions from the given app, in JSON format

        :param int application_id: Application ID

        :returns: JSON string
        """

        return self.request('/controller/actions/{0}'.format(application_id), {}, 'GET', query=True, use_json=False)

    def import_actions(self, application_id, json):
        """
        Imports all actions into the given app, from JSON format

        :param int application_id: Application ID
        :param string json: Output of export_actions

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/actions/{0}'.format(application_id)

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('actions.json', json)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def export_email_action_templates(self):
        """
        Exports all email action templates from the controller, in JSON format

        :returns: JSON string
        """

        return self.request('/controller/actiontemplate/email', {}, 'GET', query=True, use_json=False)

    def import_email_action_templates(self, json):
        """
        Imports all email action templates into the controller, from JSON format

        :param string json: Output of export_email_action_templates

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/actiontemplate/email'

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('actions.json', json)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def export_httprequest_action_templates(self):
        """
        Exports all httprequest action templates from the controller, in JSON format

        :returns: JSON string
        """

        return self.request('/controller/actiontemplate/httprequest', {}, 'GET', query=True, use_json=False)

    def import_httprequest_action_templates(self, json):
        """
        Imports all httprequest action templates into the controller, from JSON format

        :param string json: Output of export_httprequest_action_templates

        :returns: JSON string, containing success or failure messages
        """

        path = '/controller/actiontemplate/httprequest'

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        files = {'file': ('actions.json', json)}
        r = self._request(method='POST', url=url, auth=self._auth, files=files)
        return r.text

    def get_config(self):
        """
        Retrieve the controller configuration.

        :returns: Configuration variables.
        :rtype: appd.model.ConfigVariables
        """
        return self._top_request(ConfigVariables, '/configuration')

    def set_config(self, name, value):
        """
        Set a controller configuration value.

        :returns: Configuration variables.
        :rtype: appd.model.ConfigVariables
        """
        param = {'name': name, 'value': value}
        return self._top_request(ConfigVariables, '/configuration', param, 'POST', query=True)

    def get_applications(self):
        """
        Get a list of all business applications.

        :returns: List of applications visible to the user.
        :rtype: appd.model.Applications
        """
        return self._top_request(Applications, '/applications')

    def mark_nodes_historical(self, nodes):
        """
        Mark nodes historical

        :param str nodes: CSV of node IDs

        :returns: Nothing on success, exception with HTTP error code on failure
        """
        params = {'application-component-node-ids': nodes}

        return self.request('/controller/rest/mark-nodes-historical', params, 'POST', query=True, use_json=False)

    # Application-level requests

    def _app_request(self, cls, path, app_id=None, params=None, method='GET', query=True, use_json=True):
        path = self._app_path(app_id, path)
        return cls.from_json(self.request(path, params, method=method, query=query, use_json=use_json))

    def get_bt_list(self, app_id=None, excluded=False):
        """
        Get the list of all registered business transactions in an application.

        :param int app_id: Application ID to retrieve the BT list for. If :const:`None`, the value stored in the
          `app_id` property will be used.
        :param bool excluded: If True, the function will return BT's that have been excluded in the AppDynamics
          UI. If False, the function will return all BT's that have not been excluded. The default is False.
        :returns: The list of registered business transactions.
        :rtype: appd.model.BusinessTransactions
        """
        return self._app_request(BusinessTransactions, '/business-transactions', app_id, {'exclude': excluded})

    def get_backends(self, app_id=None):
        """
        Get the list of all backends in an application.

        :param int app_id: Application ID to retrieve backends for. If :const:`None`, the value stored in the
          `app_id` property will be used.
        :return: A :class:`Backends <appd.model.Backends>` object, representing a collection of backends.
        :rtype: appd.model.Backends
        """
        return self._app_request(Backends, '/backends', app_id)

    def get_tiers(self, app_id=None):
        """
        Get the list of all configured tiers in an application.

        :param int app_id: Application ID to retrieve tiers for. If :const:`None`, the value stored in the
          `app_id` property will be used.
        :return: A :class:`Tiers <appd.model.Tiers>` object, representing a collection of tiers.
        :rtype: appd.model.Tiers
        """
        return self._app_request(Tiers, '/tiers', app_id)

    def get_nodes(self, app_id=None, tier_id=None):
        """
        Retrieves the list of nodes in the application, optionally filtered by tier.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
          `app_id` property will be used.
        :param int tier_id: If set, retrieve only the nodes belonging to the specified tier. If :const:`None`,
          retrieve all nodes in the application.
        :return: A :class:`Nodes <appd.model.Nodes>` object, representing a collection of nodes.
        :rtype: appd.model.Nodes
        """

        path = ('/tiers/%s/nodes' % tier_id) if tier_id else '/nodes'
        return self._app_request(Nodes, path, app_id)

    def get_node(self, node_id, app_id=None):
        """
        Retrieves details about a single node.

        :param node_id: ID or name of the node to retrieve.
        :param app_id: Application ID to search for the node.
        :return: A single Node object.
        :rtype: appd.model.Node
        """
        return self._app_request(Node, '/nodes/%s' % node_id, app_id)

    def _validate_time_range(self, time_range_type, duration_in_mins, start_time, end_time):

        """
        Validates the combination of parameters used to specify a time range in AppDynamics.

        :param str time_range_type: type of time range to search
        :param int duration_in_mins: duration to search, in minutes
        :param long start_time: starting time
        :param long end_time: ending time
        :returns: parameters to be sent to controller
        :rtype: dict
        """
        if time_range_type and time_range_type not in self.TIME_RANGE_TYPES:
            raise ValueError('time_range_time must be one of: ' + ', '.join(self.TIME_RANGE_TYPES))

        elif time_range_type == 'BEFORE_NOW' and not duration_in_mins:
            raise ValueError('when using BEFORE_NOW, you must specify duration_in_mins')

        elif time_range_type == 'BEFORE_TIME' and (not end_time or not duration_in_mins):
            raise ValueError('when using BEFORE_TIME, you must specify duration_in_mins and end_time')

        elif time_range_type == 'AFTER_TIME' and (not start_time or not duration_in_mins):
            raise ValueError('when using AFTER_TIME, you must specify duration_in_mins and start_time')

        elif time_range_type == 'BETWEEN_TIMES' and (not start_time or not end_time):
            raise ValueError('when using BETWEEN_TIMES, you must specify start_time and end_time')

        return {'time-range-type': time_range_type,
                'duration-in-mins': duration_in_mins,
                'start-time': start_time,
                'end-time': end_time}

    def get_metrics(self, metric_path, app_id=None, time_range_type='BEFORE_NOW',
                    duration_in_mins=15, start_time=None, end_time=None, rollup=True):
        """
        Retrieves metric data.

        :param str metric_path: Full metric path of the metric(s) to be retrieved. Wildcards are supported.
            See :ref:`metric-paths` for details.
        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str time_range_type: Must be one of :const:`BEFORE_NOW`, :const:`BEFORE_TIME`,
            :const:`AFTER_TIME`, or :const:`BETWEEN_TIMES`.
            See :ref:`time-range-types` for a full explanation.
        :param int duration_in_mins: Number of minutes before now. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_NOW`.
        :param long start_time: Start time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`AFTER_TIME` or :const:`BETWEEN_TIMES`.
        :param long end_time: End time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_TIME` or :const:`BETWEEN_TIMES`.
        :param bool rollup: If :const:`False`, return individual data points for each time slice in
            the given time range. If :const:`True`, aggregates the data and returns a single data point.
        :returns: A list of metric values.
        :rtype: appd.model.MetricData
        """

        params = self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)
        params.update({'metric-path': metric_path,
                       'rollup': rollup,
                       'output': 'JSON'})

        return self._app_request(MetricData, '/metric-data', app_id, params)

    def get_events(self, app_id=None, event_types='ALL', severities='ERROR', time_range_type='BEFORE_NOW',
                   duration_in_mins=15, start_time=None, end_time=None):
        """
        Retrieves events.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str event_types: CSV of the events types to search on
        :param str severities: CSV of the severities to search on
        :param str time_range_type: Must be one of :const:`BEFORE_NOW`, :const:`BEFORE_TIME`,
            :const:`AFTER_TIME`, or :const:`BETWEEN_TIMES`.
            See :ref:`time-range-types` for a full explanation.
        :param int duration_in_mins: Number of minutes before now. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_NOW`.
        :param long start_time: Start time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`AFTER_TIME` or :const:`BETWEEN_TIMES`.
        :param long end_time: End time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_TIME` or :const:`BETWEEN_TIMES`.
        :param bool rollup: If :const:`False`, return individual data points for each time slice in
            the given time range. If :const:`True`, aggregates the data and returns a single data point.
        :returns: A list of events for the specified app, filtered by the event type and severities desired.
        :rtype: appd.model.Events
        """

        params = self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)
        params.update({'event-types': event_types,
                       'severities': severities,
                       'output': 'JSON'})

        return self._app_request(Events, '/events', app_id, params)

    def create_event(self, app_id=None, summary='', comment='', eventtype='APPLICATION_CONFIG_CHANGE',
                     severity='INFO', customeventtype=None, node=None, tier=None, bt=None):
        """
        Creates an event.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str summary: Summary of the event
        :param str comment: Comment about the event
        :param str eventtype: Event type
        :param str severity: Severity of event, one of INFO, WARN, ERROR
        :returns: Message containing the event ID if successful
        """

        # TODO: support array of properties

        params = {'summary': summary,
                  'comment': comment,
                  'eventtype': eventtype,
                  'severity': severity}

        if (eventtype == "CUSTOM"):
            params.update({'customeventtype': customeventtype,
                           'node': node,
                           'tier': tier,
                           'bt': bt})

        path = self._app_path(app_id, '/events')
        return self.request(path, params, method='POST', query=True, use_json=False)

    def exclude_bt_list(self, app_id, bts=[], exclude=True):
        xml = '<business-transactions>'

        for bt_id in bts:
            xml = xml + '<business-transaction>'
            xml = xml + '<id>{0}</id>'.format(bt_id)
            xml = xml + '</business-transaction>'

        xml = xml + '</business-transactions>'

        params = {}
        if (exclude):
            params.update({'exclude': 'true'})
        else:
            params.update({'exclude': 'false'})

        path = self._app_path(app_id, '/business-transactions')

        if not path.startswith('/'):
            path = '/' + path
        url = self._base_url + path

        headers = {'Content-type': 'text/xml', 'Accept': 'text/plain'}
        r = self._request(method='POST', url=url, auth=self._auth, params=params, data=xml, headers=headers)

        return r

    def get_snapshots(self, app_id=None, time_range_type=None, duration_in_mins=None,
                      start_time=None, end_time=None, **kwargs):
        """
        Finds and returns any snapshots in the given time range that match a set of criteria. You must provide
        at least one condition to the search parameters in the :data:`kwargs` parameters. The list of valid
        conditions can be found `here <http://appd.ws/2>`_.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str time_range_type: Must be one of :const:`BEFORE_NOW`, :const:`BEFORE_TIME`,
            :const:`AFTER_TIME`, or :const:`BETWEEN_TIMES`.
            See :ref:`time-range-types` for a full explanation.
        :param int duration_in_mins: Number of minutes before now. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_NOW`.
        :param long start_time: Start time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`AFTER_TIME` or :const:`BETWEEN_TIMES`.
        :param long end_time: End time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_TIME` or :const:`BETWEEN_TIMES`.
        :param kwargs: Additional key/value pairs to pass to the controller as search parameters.
        :returns: A list of snapshots.
        :rtype: appd.model.Snapshots
        """

        self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)

        params = self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)

        for qs_name in self.SNAPSHOT_REQUEST_PARAMS:
            arg_name = qs_name.replace('-', '_')
            params[qs_name] = kwargs.get(arg_name, None)
            if qs_name in self.SNAPSHOT_REQUEST_LISTS and qs_name in kwargs:
                params[qs_name] = ','.join(params[qs_name])

        return self._app_request(Snapshots, '/request-snapshots', app_id, params)

    def get_policy_violations(self, app_id=None, time_range_type='BEFORE_NOW', duration_in_mins=15,
                              start_time=None, end_time=None):
        """
        Retrieves a list of policy violations during the specified time range.
        *NOTE:* Beginning with controller version 3.7, you should use :meth:`get_healthrule_violations` instead.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str time_range_type: Must be one of :const:`BEFORE_NOW`, :const:`BEFORE_TIME`,
            :const:`AFTER_TIME`, or :const:`BETWEEN_TIMES`.
            See :ref:`time-range-types` for a full explanation.
        :param int duration_in_mins: Number of minutes before now. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_NOW`.
        :param long start_time: Start time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`AFTER_TIME` or :const:`BETWEEN_TIMES`.
        :param long end_time: End time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_TIME` or :const:`BETWEEN_TIMES`.
        :returns: A list of policy violations.
        :rtype: appd.model.PolicyViolations
        """

        params = self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)

        return self._app_request(PolicyViolations, '/problems/policy-violations', app_id, params)

    def get_healthrule_violations(self, app_id=None, time_range_type='BEFORE_NOW', duration_in_mins=15,
                                  start_time=None, end_time=None):
        """
        Retrieves a list of health rule violations during the specified time range. Compatible with
        controller version 3.7 and later.

        :param int app_id: Application ID to retrieve nodes for. If :const:`None`, the value stored in the
            `app_id` property will be used.
        :param str time_range_type: Must be one of :const:`BEFORE_NOW`, :const:`BEFORE_TIME`,
            :const:`AFTER_TIME`, or :const:`BETWEEN_TIMES`.
            See :ref:`time-range-types` for a full explanation.
        :param int duration_in_mins: Number of minutes before now. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_NOW`.
        :param long start_time: Start time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`AFTER_TIME` or :const:`BETWEEN_TIMES`.
        :param long end_time: End time, expressed in milliseconds since epoch. Only valid if the
            :attr:`time_range_type` is :const:`BEFORE_TIME` or :const:`BETWEEN_TIMES`.
        :returns: A list of policy violations.
        :rtype: appd.model.PolicyViolations
        """

        params = self._validate_time_range(time_range_type, duration_in_mins, start_time, end_time)

        return self._app_request(PolicyViolations, '/problems/healthrule-violations', app_id, params)

    def _v2_request(self, cls, path, params=None, method='GET'):
        return cls.from_json(self.request('/api' + path, params, method))

    def get_my_account(self):
        """
        :rtype: Account
        """
        return self._v2_request(Account, '/accounts/myaccount')

    def get_account(self, account_id):
        """
        :rtype: Account
        """
        return self._v2_request(Account, '/accounts/{0}'.format(account_id))

    def get_license_modules(self, account_id):
        """
        :rtype: LicenseModules
        """
        return self._v2_request(LicenseModules, '/accounts/{0}/licensemodules'.format(account_id))

    def set_metric_retention(self, account_id, days, app_id=None):
        path = '/controller/api/accounts/{0}/'.format(account_id)

        if (app_id):
            path = path + 'applications/{0}/'.format(app_id)

        path = path + 'metricstaleduration/{0}'.format(days)

        return self.request(path, method="POST", use_json=False)

    def get_action_suppressions(self, account_id, app_id):
        return self._v2_request(ActionSuppressionsResponse,
                                '/accounts/{0}/applications/{1}/actionsuppressions'.format(account_id, app_id))

    def get_action_suppression(self, account_id, app_id, action_suppression_id):
        return self._v2_request(ActionSuppression,
                                '/accounts/{0}/applications/{1}/actionsuppressions/{2}'.format(account_id,
                                                                                               app_id,
                                                                                               action_suppression_id))

    def delete_action_suppression(self, account_id, app_id, action_suppression_id):
        return self.request('/api/accounts/{0}/applications/{1}/actionsuppressions/{2}'.format(account_id,
                                                                                               app_id,
                                                                                               action_suppression_id),
                            method="DELETE",
                            use_json=False)

    def create_action_suppression(self, account_id, app_id, params):
        return self.request('/api/accounts/{0}/applications/{1}/actionsuppressions'.format(account_id, app_id),
                            method="POST",
                            use_json=False,
                            query=False,
                            params=params,
                            headers={'Content-Type': 'application/vnd.appd.cntrl+json;v=1'})

    def get_license_usage(self, account_id, license_module=None, start_time=None, end_time=None):
        """
        :param int account_id:
        :param str license_module:
        :param datetime.datetime start_time:
        :param datetime.datetime end_time:
        :rtype: HourlyLicenseUsages
        """
        if isinstance(start_time, datetime):
            start_time = start_time.isoformat()
        if isinstance(end_time, datetime):
            end_time = end_time.isoformat()

        params = {
            'licensemodule': license_module,
            'showfiveminutesresolution': 'False',
            'startdate': start_time,
            'enddate': end_time
        }
        return self._v2_request(HourlyLicenseUsages, '/accounts/{0}/licensemodules/usages'.format(account_id), params)

    def get_license_usage_5min(self, account_id, license_module=None, start_time=None, end_time=None):
        """
        :param int account_id:
        :param str license_module:
        :param datetime.datetime start_time:
        :param datetime.datetime end_time:
        :rtype: LicenseUsages
        """
        params = {
            'licensemodule': license_module,
            'showfiveminutesresolution': 'True',
            'startdate': start_time.isoformat() if start_time else None,
            'enddate': end_time.isoformat() if end_time else None
        }
        return self._v2_request(LicenseUsages, '/accounts/{0}/licensemodules/usages'.format(account_id), params)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import itertools
import csv

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


def incr(d, name, amt=1):
    d[name] = d.get(name, 0) + amt


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

nodes = []
for app in c.get_applications():
    for node in c.get_nodes(app.id):

        node.app_id = app.id
        node.app_name = app.name

        if node.has_machine_agent or node.has_app_agent:
            if node.has_app_agent:
                if 'PHP' in node.type:
                    node.group_type = 'PHP App Agent'
                if 'IIS' in node.type:
                    node.group_type = '.NET App Agent'
                else:
                    node.group_type = 'Java App Agent'
            else:
                node.group_type = 'Machine Agent only'

            nodes.append(node)


# Sort and group the nodes by machine_id.

group_func = lambda x: x.machine_id
nodes.sort(key=group_func)

with open('data/licenses_by_host.csv', 'w') as f:
    w = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC, dialect=csv.excel)

    hdr = ['App Name', 'Host Name', 'Host Type', 'License Count']
    w.writerow(hdr)

    for machine_id, nodes_on_machine_iter in itertools.groupby(nodes, key=group_func):

        nodes_on_machine = list(nodes_on_machine_iter)
        first_node = nodes_on_machine[0]
        agent_type = first_node.group_type
        types = [x.group_type for x in nodes_on_machine]
        all_same = all(x.group_type == agent_type for x in nodes_on_machine)
        assert all_same, first_node

        license_count = 1
        if 'Java' in agent_type:
            license_count = len(nodes_on_machine)

        w.writerow([first_node.app_name, first_node.machine_name, agent_type, license_count])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to print a list of all external dependencies for an application.
This would include databases, message queues, web sites,
Output format: ``app_name,tier_name,node_name,host_name``
"""

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

app_name = 'E-Commerce_Demo'
tier_name = 'ECommerce-Server'
metric_path = 'Overall Application Performance|' + tier_name + '|External Calls'

app_id = -1
for app in c.get_applications():
    if app.name == app_name:
        app_id = app.id

deps = c.get_metric_tree(app_id, metric_path)
for dep in deps:
    if dep.type == 'folder':
        print(dep.name)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList
from appd.time import from_ts


class Snapshot(JsonObject):

    FIELDS = {'id': '', 'local_id': 'localID', 'request_guid': 'requestGUID', 'summary': '',
              'bt_id': 'businessTransactionId', 'app_id': 'applicationId',
              'url': 'URL', 'archived': '', 'async': '', 'stall_dump': 'stallDump',
              'call_chain': 'callChain', 'is_first_in_chain': 'firstInChain',
              'diag_session_guid': 'diagnosticSessionGUID', 'exit_sequence': 'snapshotExitSequence',
              'exit_calls_truncated': 'exitCallsDataTruncated',
              'exit_calls_truncated_msg': 'exitCallsDataTruncationMessage',
              'app_component_id': 'applicationComponentId', 'app_component_node_id': 'applicationComponentNodeId',
              'local_start_time_ms': 'localStartTime', 'server_start_time_ms': 'serverStartTime',
              'thread_id': 'threadID', 'thread_name': 'threadName',
              'http_headers': 'httpHeaders', 'response_headers': 'responseHeaders',
              'http_params': 'httpParameters', 'cookies': '',
              'http_session_id': 'httpSessionID', 'session_keys': 'sessionKeys', 'business_data': 'businessData',
              'error_ids': 'errorIDs', 'error_occurred': 'errorOccured', 'error_summary': 'errorSummary',
              'error_details': 'errorDetails', 'log_messages': 'logMessages',
              'bt_events': 'transactionEvents', 'bt_properties': 'transactionProperties',
              'dotnet_properties': 'dotnetProperty',
              'has_deep_dive_data': 'hasDeepDiveData', 'deep_dive_policy': 'deepDivePolicy',
              'is_delayed_deep_dive': 'delayedDeepDive', 'delayed_deep_dive_offset': 'delayedDeepDiveOffSet',
              'has_unresolved_calls': 'unresolvedCallInCallChain',
              'time_taken_ms': 'timeTakenInMilliSecs', 'cpu_time_taken_ms': 'cpuTimeTakenInMilliSecs',
              'warning_threshold': 'warningThreshold', 'critical_threshold': 'criticalThreshold',
              'user_experience': 'userExperience'}

    def __init__(self, snap_id=0, **kwargs):
        self.id = snap_id
        self.local_start_time_ms, self.server_start_time_ms = 0, 0
        for k, v in list(Snapshot.FIELDS.items()):
            self.__setattr__(k, kwargs.get(k, None))

    @property
    def local_start_time(self):
        return from_ts(self.local_start_time_ms)

    @property
    def server_start_time(self):
        return from_ts(self.server_start_time_ms)


class Snapshots(JsonList):

    def __init__(self, initial_list=None):
        super(Snapshots, self).__init__(Snapshot, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Snapshot
        """
        return self.data[i]



"""
This module contains the main classes for handling requests to the AppDynamics REST API.
"""

from __future__ import absolute_import

from datetime import datetime
from time import mktime


def from_ts(ms):
    """
    Converts a timestamp from AppDynamics internal format to a Python :class:`datetime <datetime.datetime>` object.

    :param long ms: Timestamp expressed as milliseconds since epoch.
    :returns: Converted value.
    :rtype: datetime
    """

    return datetime.fromtimestamp(ms / 1000)


def to_ts(dt):
    """
    Converts a timestamp from Python :class:`datetime <datetime.datetime>` to AppDynamics format.

    :param datetime dt: Timestamp to convert.
    :returns: Converted value.
    :rtype: long
    """

    return int(mktime(dt.timetuple())) * 1000





#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to generate a list of all health rule violations that occurred yesterday.
"""

from __future__ import print_function

from datetime import datetime

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient
from appd.time import from_ts, to_ts


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


# Parse the command line arguments and initialize the client
#
args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

# Get the application list and find "MyApp"
#
apps = c.get_applications()
gn_prod = [x for x in apps if x.name == "MyApp"][0]

# Calculate start and end times for the report: 24h period ending with midnight last night
#
today = to_ts(datetime.now().date())
yesterday = today - (86400 * 1000)

# Get the list of nodes so we can look up by node_id
#
all_nodes = c.get_nodes(gn_prod.id)
nodes_by_id = dict(list(zip([x.id for x in all_nodes], all_nodes)))

# Get all health rule violations and print them
#
violations = c.get_healthrule_violations(gn_prod.id, 'BETWEEN_TIMES', start_time=yesterday, end_time=today)
for v in violations:
    print("-" * 70)
    print("Start time:   ", from_ts(v.start_time_ms))
    print("End time:     ", from_ts(v.end_time_ms))
    print("Description:  ", v.description)
    print("Details:      ", v.deep_link_url)
    if v.affected_entity.type == "APPLICATION_COMPONENT_NODE":
        try:
            print("Affected Node: ", nodes_by_id[v.affected_entity.entity_id].name)
        except KeyError:
            print("Affected Node: UNKNOWN")



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for AppDynamics REST API
"""

import unittest
from test import ApplicationApiTest
from datetime import datetime, timedelta
import creds.demo2 as creds
import appd
import tzlocal


def now():
    mytz = tzlocal.get_localzone()
    return datetime.now(mytz).replace(microsecond=0)


def midnight():
    return now().replace(hour=0, minute=0, second=0, microsecond=0)


class V2_LicenseUsageTest(ApplicationApiTest):

    def setUp(self):
        self.c = appd.request.AppDynamicsClient(creds.url, creds.user, creds.password, creds.account)

    def test_license_usage_java(self):
        usage = self.c.get_license_usage(2, 'java', midnight() - timedelta(hours=24), midnight() - timedelta(hours=1))
        for x in usage.usages:
            self.assertEqual(x.license_module, 'java')
            self.assertEqual(x.account_id, 2)
            self.assertEqual(x.sample_count, 12)
            self.assertEqual(x.avg_units_provisioned, 1999)
            self.assertEqual(x.avg_units_allowed, 1999)
        self.assertEqual(len(usage.usages), 24, 'License usage should return 24 data points')

    def test_license_usage_java_5min(self):
        usage = self.c.get_license_usage_5min(2, 'java', now() - timedelta(hours=1), now())
        for x in usage.usages:
            self.assertEqual(x.license_module, 'java')
            self.assertEqual(x.account_id, 2)
            self.assertEqual(x.units_allowed, 1999)
            self.assertEqual(x.units_provisioned, 1999)
        self.assertEqual(len(usage.usages), 12, 'License usage should return 12 data points')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for AppDynamics REST API
"""

import unittest
from test import ApplicationApiTest


class V2_LicenseModuleTest(ApplicationApiTest):

    def test_license_modules(self):
        mods = self.c.get_license_modules(2)
        self.assertIn('java', mods.modules)
        self.assertIn('dot-net', mods.modules)
        self.assertNotIn('wharrgarbl', mods.modules)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(V2_LicenseModuleTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from collections import OrderedDict, defaultdict

import pygal
import tzlocal

from appd.request import AppDynamicsClient
import creds.demo2 as creds

__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'

# Set up API client
c = AppDynamicsClient(creds.url, creds.user, creds.password, creds.account)

# Get my tenant account info
my_acct = c.get_my_account()

# Calculate start and end dates - we will start at midnight last night and go back 7 days
days = 15
mytz = tzlocal.get_localzone()
end_dt = datetime.now(mytz).replace(hour=0, minute=0, second=0, microsecond=0)
start_dt = end_dt - timedelta(days)

# Get license usage for my account
usage = c.get_license_usage(my_acct.id, None, start_dt, end_dt)


def daterange(start_dt, end_dt):
    for n in range(int((end_dt - start_dt).days)):
        yield start_dt + timedelta(days=n)

# Get the list of all licensed products:
products = set(x.license_module for x in usage.usages)
products = [x for x in products if 'eum' not in x and 'analytics' not in x]

usage_by_product = defaultdict(OrderedDict)
for product in products:
    for day in daterange(start_dt, end_dt):
        units = [x.max_units_used for x in usage.usages if
                 x.created_on.date() == day.date() and x.license_module == product]
        usage_by_product[product][day] = max(units)

# Make a simple graph and display it
chart = pygal.StackedBar(x_label_rotation=45, width=1000)
chart.title = 'License Usage By Product - ' + c.base_url
chart.x_labels = [str(x.date()) for x in daterange(start_dt, end_dt)]
for product in products:
    chart.add(product, usage_by_product[product].values())
chart.render_to_file('all_usage.svg')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for AppDynamics REST API
"""

from test import ApplicationApiTest

class V1_ApplicationTest(ApplicationApiTest):

    def test_app_list(self):
        apps = self.c.get_applications()
        self.assertEqual(len(apps), 1)
        self.assertEqual(apps[0].id, 10)
        self.assertEqual(apps[0].name, 'ACME Book Store Application')

    def test_get_app_by_name(self):
        apps = self.c.get_applications()
        self.assertEqual(apps.by_name('ACME Book Store Application').id, 10)

    def test_get_config_by_name(self):
        config = self.c.get_config()
        val = config.by_name('metrics.retention.period')
        self.assertEqual(val.name, 'metrics.retention.period')
        self.assertEqual(val.value, '365')



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Kyle Furlong <kyle.furlong@appdynamics.com>
"""

from . import JsonObject, JsonList


class Audit(JsonObject):

    FIELDS = {'timestamp': 'timeStamp', 'account_name': 'accountName',
              'security_provider_type': 'securityProviderType',
              'user_name': 'userName', 'action': 'action'}

    def __init__(self, timestamp=0, account_name=None, security_provider_type=None, user_name=None, action=None):
        (self.timestamp, self.account_name, self.security_provider_type,
         self.user_name, self.action) = (timestamp, account_name, security_provider_type, user_name, action)


class AuditHistory(JsonList):

    def __init__(self, initial_list=None):
        super(AuditHistory, self).__init__(Audit, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Node
        """
        return self.data[i]

    def by_action(self, action):
        """
        Filters an AuditHistory collection to return only the Audits matching the given action.
        :param str name: Action to match against.
        :returns: a AuditHistory collection filtered by action.
        :rtype: AuditHistory
        """
        return AuditHistory([x for x in self.data if x.action == action])

    def by_user_name(self, user_name):
        """
        Filters an AuditHistory collection to return only the Audits matching the given user name.
        :param str name: User name to match against.
        :returns: a AuditHistory collection filtered by user name.
        :rtype: AuditHistory
        """
        return AuditHistory([x for x in self.data if x.user_name == user_name])

    def by_account_name(self, account_name):
        """
        Filters an AuditHistory collection to return only the Audits matching the given account.
        :param str name: Account to match against.
        :returns: a AuditHistory collection filtered by account.
        :rtype: AuditHistory
        """
        return AuditHistory([x for x in self.data if x.account_name == account_name])



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""


from future.moves.collections import UserList


def _filter_func(obj, pred):
    def func():
        return obj.__class__([item for item in obj.data])

    return func


class JsonObject(object):

    FIELDS = {}

    @classmethod
    def _set_fields_from_json_dict(cls, obj, json_dict):
        for k, v in list(obj.FIELDS.items()):
            obj.__setattr__(k, json_dict[v or k])

    @classmethod
    def from_json(cls, json_dict):
        obj = cls()
        cls._set_fields_from_json_dict(obj, json_dict)
        return obj

    def __str__(self):
        rep = ', '.join([x + '=' + repr(y) for x, y in list(self.__dict__.items())])
        return '<{0}: {1}>'.format(self.__class__.__name__, rep)

    __repr__ = __str__

    def _list_setter(self, attr_name, new_val, allowed_vals):
        if new_val and (new_val not in allowed_vals):
            raise ValueError('{0} must be one of [{1}] but got {2}'.format(
                attr_name, ', '.join(allowed_vals), new_val))
        self.__setattr__(attr_name, new_val)


class JsonList(UserList):

    OBJECT_TYPE = None

    def __init__(self, cls, initial_list=None):
        UserList.__init__(self)
        if initial_list:
            for item in initial_list:
                if isinstance(item, cls):
                    self.data.append(item)
                elif isinstance(item, dict):
                    self.data.append(cls.from_json(item))

    @classmethod
    def from_json(cls, json_list):
        return cls(json_list)

    def __str__(self):
        rep = ', '.join([str(x) for x in self.data])
        return '<{0}[{1}]: {2}>'.format(self.__class__.__name__, len(self.data), rep)



"""
Model classes for AppDynamics REST API
"""

from . import JsonObject, JsonList
from appd.time import from_ts

class LicenseUsage(JsonObject):

    FIELDS = {
        'id': '',
        'account_id': 'accountId',
        'units_used': 'unitsUsed',
        'units_allowed': 'unitsAllowed',
        'units_provisioned': 'unitsProvisioned',
        'license_module': 'agentType',
        'created_on_ms': 'createdOn',
        }

    def __init__(self, id=0, account_id=0, license_module=None, units_used=0,
                 units_allowed=0, units_provisioned=None, created_on_ms=0):
        (self.id, self.account_id, self.license_module, self.units_used,
         self.units_allowed, self.units_provisioned, self.created_on_ms) = (id, account_id, license_module,
                                                                            units_used, units_allowed,
                                                                            units_provisioned, created_on_ms)

    @property
    def created_on(self):
        """
        :rtype: datetime.datetime
        """
        return from_ts(self.created_on_ms)


class LicenseUsageList(JsonList):

    def __init__(self, initial_list=None):
        super(LicenseUsageList, self).__init__(LicenseUsage, initial_list)

    def __getitem__(self, i):
        """
        :rtype: LicenseUsage
        """
        return self.data[i]

    def by_account_id(self, account_id):
        """
        :rtype: LicenseUsageList
        """
        return LicenseUsageList([x for x in self.data if x.account_id == account_id])

    def by_license_module(self, license_module):
        """
        :rtype: LicenseUsageList
        """
        return LicenseUsageList([x for x in self.data if x.license_module == license_module])


class LicenseUsages(JsonObject):

    FIELDS = {}

    def __init__(self):
        self.usages = LicenseUsageList()

    @classmethod
    def from_json(cls, json_dict):
        obj = super(LicenseUsages, cls).from_json(json_dict)
        obj.usages = LicenseUsageList.from_json(json_dict['usages'])
        return obj




#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient

__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

for app in c.get_applications():
    metric_data = c.get_metrics('Overall Application Performance|*', app_id=app.id)
    art = metric_data.by_leaf_name(c.AVERAGE_RESPONSE_TIME).first_value()
    cpm = metric_data.by_leaf_name(c.CALLS_PER_MINUTE).first_value()
    epm = metric_data.by_leaf_name(c.ERRORS_PER_MINUTE).first_value()
    error_pct = round(float(epm) / float(cpm) * 100.0, 1) if cpm > 0 else 0
    print(app.name, art, cpm, epm, error_pct)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to generate a count of licenses used by app name, regardless of environment type
(Prod, Devl, Qual, Cert, etc.)
"""

from __future__ import print_function

from datetime import datetime
import itertools

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


def incr(d, name, amt=1):
    d[name] = d.get(name, 0) + amt


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

nodes = []
for app in c.get_applications():
    for node in c.get_nodes(app.id):
        if node.has_machine_agent or node.has_app_agent:
            if node.has_app_agent:
                if 'PHP' in node.type:
                    node.group_type = 'PHP App Agent'
                if 'IIS' in node.type:
                    node.group_type = '.NET App Agent'
                else:
                    node.group_type = 'Java App Agent'
            else:
                node.group_type = 'Machine Agent only'
            node.app = app
            nodes.append(node)


# Sort and group the nodes by machine_id.

group_func = lambda x: x.machine_id
nodes.sort(key=group_func)

host_counts = dict()
node_counts = dict()
lic_counts = dict()
for machine_id, nodes_on_machine_iter in itertools.groupby(nodes, key=group_func):

    nodes_on_machine = list(nodes_on_machine_iter)
    first_node = nodes_on_machine[0]
    agent_type = first_node.group_type

    app_name = first_node.app.name
    app_name = app_name.replace('Prod', '').replace('Qual', '').replace('Cert', '').replace('Devl', '')

    all_same = all(x.group_type == agent_type for x in nodes_on_machine)
    # assert all_same, first_node

    all_same = all(x.app.name == app_name for x in nodes_on_machine)
    # assert all_same, first_node

    license_count = 1
    if 'Java' in agent_type:
        license_count = len(nodes_on_machine)

    incr(lic_counts, app_name, license_count)
    incr(host_counts, app_name, 1)
    incr(node_counts, app_name, len(nodes_on_machine))


# Print the results.
tot_nodes, tot_hosts, tot_licenses = (0, 0, 0)
header_fmt = '%-30s\t%-15s\t%-15s\t%s'
data_fmt = '%-30s\t%15d\t%15d\t%15d'

print()
print('License usage report for ', args.url)
print('Generated at: ', datetime.now())
print()
print(header_fmt % ('App Name', 'Node Count', 'Host Count', 'License Count'))
print(header_fmt % ('=' * 30, '=' * 15, '=' * 15, '=' * 15))

for app_name in sorted(node_counts.keys()):
    node_count = node_counts.get(app_name, 0)
    host_count = host_counts.get(app_name, 0)
    lic_count = lic_counts.get(app_name, 0)
    tot_nodes += node_count
    tot_hosts += host_count
    tot_licenses += lic_count
    print(data_fmt % (app_name, node_count, host_count, lic_count))

print(header_fmt % ('=' * 30, '=' * 15, '=' * 15, '=' * 15))
print(data_fmt % ('TOTAL', tot_nodes, tot_hosts, tot_licenses))



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList
from .entity_def import EntityDefinition
from appd.time import from_ts


class PolicyViolation(JsonObject):

    FIELDS = {'id': '',
              'name': '',
              'description': '',
              'status': 'incidentStatus',
              'severity': '',
              'start_time_ms': 'startTimeInMillis',
              'end_time_ms': 'endTimeInMillis',
              'detected_time_ms': 'detectedTimeInMillis',
              'deep_link_url': 'deepLinkUrl'}

    STATUSES = ('NOT_APPLICABLE', 'OPEN', 'RESOLVED')

    SEVERITIES = ('INFO', 'WARNING', 'CRITICAL')

    def __init__(self, pv_id=0, pv_name='', description='', status='OPEN', severity='INFO',
                 start_time_ms=0, end_time_ms=0, detected_time_ms=0, deep_link_url='',
                 affected_entity=None, triggered_entity=None):
        self._severity = None
        self._status = None
        (self.id, self.name, self.description, self.status, self.severity, self.start_time_ms, self.end_time_ms,
         self.detected_time_ms, self.deep_link_url, self.affected_entity, self.triggered_entity) = \
            (pv_id, pv_name, description, status, severity, start_time_ms, end_time_ms, detected_time_ms,
             deep_link_url, affected_entity, triggered_entity)

    @classmethod
    def _set_fields_from_json_dict(cls, obj, json_dict):
        JsonObject._set_fields_from_json_dict(obj, json_dict)
        obj.affected_entity = EntityDefinition.from_json(json_dict['affectedEntityDefinition'])
        obj.triggered_entity = EntityDefinition.from_json(json_dict['triggeredEntityDefinition'])

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, new_status):
        self._list_setter('_status', new_status, PolicyViolation.STATUSES)

    @property
    def severity(self):
        return self._severity

    @severity.setter
    def severity(self, new_sev):
        self._list_setter('_severity', new_sev, PolicyViolation.SEVERITIES)

    @property
    def start_time(self):
        return from_ts(self.start_time_ms)

    @property
    def end_time(self):
        """
        Gets the end time of the violation, converting it from an AppDynamics timestamp to standard
        Python :class:datetime.

        :return: Time the violation was resolved
        :rtype: datetime
        """
        return from_ts(self.end_time_ms)

    @property
    def detected_time(self):
        return from_ts(self.detected_time_ms)


class PolicyViolations(JsonList):

    def __init__(self, initial_list=None):
        super(PolicyViolations, self).__init__(PolicyViolation, initial_list)

    def __getitem__(self, i):
        """
        :rtype: PolicyViolation
        """
        return self.data[i]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import itertools
import csv

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

nodes = []
for app in c.get_applications():
    for node in c.get_nodes(app.id):
        node_type = node.type
        if node.has_machine_agent and not node.has_app_agent:
            node.type = 'Machine Agent'
        nodes.append(node)


# Sort and group the nodes by machine_id.

group_func = lambda x: (x.machine_id, x.type)
nodes.sort(key=group_func)

tier_names = set()
tiers = dict()
for machine_id, nodes_on_machine_iter in itertools.groupby(nodes, key=group_func):

    nodes_on_machine = list(nodes_on_machine_iter)
    agent_type = nodes_on_machine[0].type
    tier_name = nodes_on_machine[0].tier_name.split('.')[0]

    license_count = len(nodes_on_machine)
    if 'PHP' in agent_type:
        agent_type = 'PHP'
    if 'IIS' in agent_type:
        agent_type = 'DOT_NET'
        license_count = 1
    elif agent_type == 'Machine Agent':
        agent_type = 'MACHINE'
        license_count = 1
        assert len(nodes_on_machine) == 1
    else:
        agent_type = 'JAVA'

    def key(name):
        return '%s|%s|%s' % (tier_name, agent_type, name)

    def incr(name, amt=1):
        k = key(name)
        tiers[k] = tiers.get(k, 0) + amt

    incr('licenses', license_count)
    incr('agents', len(nodes_on_machine))
    tiers.setdefault(key('host_set'), set()).add(machine_id[0])
    tiers[key('hosts')] = len(tiers[key('host_set')])
    tier_names.add(tier_name)


AGENT_TYPES = ('JAVA', 'DOT_NET', 'PHP', 'MACHINE')

with open('data/licenses.csv', 'w') as f:
    w = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC, dialect=csv.excel)

    hdr = ['Tier Name', 'Java Licenses', 'Java Agents', 'Java Hosts', '.NET Licenses', '.NET Agents', '.NET Hosts',
           'PHP Licenses', 'PHP Agents', 'PHP Hosts', 'Machine Agent Licenses', 'Machine Agents', 'Machine Agent Hosts']
    w.writerow(hdr)

    for tier_name in sorted(tier_names):
        row = [tier_name]
        for agent_type in AGENT_TYPES:
            def get(name):
                return tiers.get('%s|%s|%s' % (tier_name, agent_type, name), 0)
            row.extend([get('licenses'), get('agents'), get('hosts')])
        w.writerow(row)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList
from .entity_def import EntityDefinition


class Event(JsonObject):

    FIELDS = {'id': '',
              'summary': '',
              'type': '',
              'archived': '',
              'deep_link_url': 'deepLinkUrl',
              'event_time_ms': 'eventTime',
              'is_read': 'markedAsRead',
              'is_resolved': 'markedAsResolved',
              'severity': '',
              'sub_type': 'subType'}

    """
    https://docs.appdynamics.com/display/PRO42/Events+Reference
    """
    EVENT_TYPES = (
        'ACTIVITY_TRACE',
        'ADJUDICATION_CANCELLED',
        'AGENT_ADD_BLACKLIST_REG_LIMIT_REACHED',
        'AGENT_ASYNC_ADD_REG_LIMIT_REACHED',
        'AGENT_CONFIGURATION_ERROR',
        'AGENT_DIAGNOSTICS',
        'AGENT_ERROR_ADD_REG_LIMIT_REACHED',
        'AGENT_EVENT',
        'AGENT_METRIC_BLACKLIST_REG_LIMIT_REACHED',
        'AGENT_METRIC_REG_LIMIT_REACHED',
        'AGENT_STATUS',
        'ALREADY_ADJUDICATED',
        'APPDYNAMICS_CONFIGURATION_WARNINGS',
        'APPDYNAMICS_DATA',
        'APPLICATION_CONFIG_CHANGE',
        'APPLICATION_DEPLOYMENT',
        'APPLICATION_ERROR',
        'APP_SERVER_RESTART',
        'AZURE_AUTO_SCALING',
        'BACKEND_DISCOVERED',
        'BT_DISCOVERED',
        'CONTROLLER_AGENT_VERSION_INCOMPATIBILITY',
        'CONTROLLER_ASYNC_ADD_REG_LIMIT_REACHED',
        'CONTROLLER_ERROR_ADD_REG_LIMIT_REACHED',
        'CONTROLLER_EVENT_UPLOAD_LIMIT_REACHED',
        'CONTROLLER_METRIC_REG_LIMIT_REACHED',
        'CONTROLLER_RSD_UPLOAD_LIMIT_REACHED',
        'CUSTOM',
        'CUSTOM_ACTION_END',
        'CUSTOM_ACTION_FAILED',
        'CUSTOM_ACTION_STARTED',
        'DEADLOCK',
        'DIAGNOSTIC_SESSION',
        'DISK_SPACE',
        'EMAIL_SENT',
        'EUM_CLOUD_BROWSER_EVENT',
        'HIGH_END_TO_END_LATENCY',
        'INFO_INSTRUMENTATION_VISIBILITY',
        'INTERNAL_UI_EVENT',
        'LICENSE',
        'LOW_HEAP_MEMORY',
        'MACHINE_DISCOVERED',
        'MEMORY',
        'MEMORY_LEAK',
        'MEMORY_LEAK_DIAGNOSTICS',
        'MOBILE_CRASH_IOS_EVENT',
        'MOBILE_CRASH_ANDROID_EVENT',
        'NODE_DISCOVERED',
        'NORMAL',
        'OBJECT_CONTENT_SUMMARY',
        'POLICY_CANCELED',
        'POLICY_CANCELED_CRITICAL',
        'POLICY_CANCELED_WARNING',
        'POLICY_CLOSE',
        'POLICY_CLOSE_CRITICAL',
        'POLICY_CLOSE_WARNING',
        'POLICY_CONTINUES',
        'POLICY_CONTINUES_CRITICAL',
        'POLICY_CONTINUES_WARNING',
        'POLICY_DOWNGRADED',
        'POLICY_OPEN',
        'POLICY_OPEN_CRITICAL',
        'POLICY_OPEN_WARNING',
        'POLICY_UPGRADED',
        'RESOURCE_POOL_LIMIT',
        'RUNBOOK_DIAGNOSTIC_SESSION_END',
        'RUNBOOK_DIAGNOSTIC_SESSION_FAILED',
        'RUNBOOK_DIAGNOSTIC_SESSION_STARTED',
        'RUN_LOCAL_SCRIPT_ACTION_END',
        'RUN_LOCAL_SCRIPT_ACTION_FAILED',
        'RUN_LOCAL_SCRIPT_ACTION_STARTED',
        'SERVICE_ENDPOINT_DISCOVERED',
        'SLOW',
        'SMS_SENT',
        'STALL',
        'STALLED',
        'SYSTEM_LOG'
        'THREAD_DUMP_ACTION_END',
        'THREAD_DUMP_ACTION_FAILED',
        'THREAD_DUMP_ACTION_STARTED',
        'TIER_DISCOVERED',
        'VERY_SLOW',
        'WORKFLOW_ACTION_END',
        'WORKFLOW_ACTION_FAILED',
        'WORKFLOW_ACTION_STARTED')

    def __init__(self, event_id=0, event_type='CUSTOM', sub_type='', summary='', archived=False, event_time_ms=0,
                 is_read=False, is_resolved=False, severity='INFO', deep_link_url='',
                 triggered_entity=None, affected_entities=None):
        self._event_type = None
        (self.id, self.type, self.sub_type, self.summary, self.archived, self.event_time_ms, self.is_read,
         self.is_resolved, self.severity, self.deep_link_url) = (event_id, event_type, sub_type, summary, archived,
                                                                 event_time_ms, is_read, is_resolved, severity,
                                                                 deep_link_url)
        self.triggered_entity = triggered_entity or EntityDefinition()
        self.affected_entities = affected_entities or []

    @property
    def event_type(self):
        """
        :return:
        """
        return self._event_type

    @event_type.setter
    def event_type(self, new_type):
        self._list_setter('_event_type', new_type, Event.EVENT_TYPES)


class Events(JsonList):

    def __init__(self, initial_list=None):
        super(Events, self).__init__(Event, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Event
        """
        return self.data[i]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to print a list of events from the specified controller and app.
"""

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient

__author__ = 'Kyle Furlong'
__copyright__ = 'Copyright (c) 2013-2017 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

account = c.get_my_account()

action = {'name': 'suppress',
          'timeRange': {'startTimeMillis': '2017-03-24T16:16:57+0000',
                        'endTimeMillis': '2017-10-25T04:16:57+0000'},
          'heathRuleIds': '1,2,3',
          'affects': {'type': 'APP'}}
resp = c.create_action_suppression(account.id, app_id=16, params=action)
print(resp)

resp = c.get_action_suppressions(account.id, app_id=16)
print(resp)

id = resp.actionSuppressions[0].id
resp = c.get_action_suppression(account.id, app_id=16, action_suppression_id=id)
print(resp)

resp = c.delete_action_suppression(account.id, app_id=16, action_suppression_id=id)
print(resp)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to print a list of events from the specified controller and app.
"""

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient

__author__ = 'Kyle Furlong'
__copyright__ = 'Copyright (c) 2013-2017 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

resp = c.export_email_action_templates()
print(resp)
resp = c.import_email_action_templates(resp)
print(resp)



"""
Command line parsing for REST API scripts.
"""

import argparse


def parse_argv(app_desc=None):
    """
    Provides a parser for a standard set of command line options.

    Use of this function is optional but highly encouraged, as it allows all scripts based on the SDK to be
    invoked in a familiar way.

    :param str app_desc: Short description of your application, to be printed when help is requested.
    :returns: a Namespace full of parsed arguments
    :rtype: argparse.Namespace
    """

    parser = argparse.ArgumentParser(description=app_desc)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Enable verbose output')
    parser.add_argument('-c', '--url', dest='url', required=True,
                        help='set the controller URL')
    parser.add_argument('-u', '--username', dest='username', required=True,
                        help='set the username for authentication')
    parser.add_argument('-p', '--password', dest='password', required=True,
                        help='set the password for authentication')
    parser.add_argument('-a', '--account', default='customer1', dest='account',
                        help='set the controller account (default: customer1)')
    return parser.parse_args()




"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class ConfigVariable(JsonObject):
    """
    Represents a controller configuration variable. The following attributes are defined:

    .. data:: name

       Variable name.

    .. data:: value

      Current value.

    .. data:: description

      Optional description of the variable.

    .. data:: updateable

      If :const:`True`, value can be changed.

    .. data:: scope

      Scope of the variable. The scope can be ``'cluster'`` or ``'local'``. Variables with cluster scope are
      replicated across HA controllers; local variables are not.
    """
    FIELDS = {
        'name': '',
        'description': '',
        'scope': '',
        'updateable': '',
        'value': ''
    }

    def __init__(self, name='', description='', scope='cluster', updateable=True, value=None):
        (self.name, self.description, self.scope, self.updateable, self.value) = (name, description, scope,
                                                                                  updateable, value)


class ConfigVariables(JsonList):
    """
    Represents a collection of :class:`ConfigVariable` objects. Extends :class:`UserList`, so it supports the
    standard array index and :keyword:`for` semantics.
    """
    def __init__(self, initial_list=None):
        super(ConfigVariables, self).__init__(ConfigVariable, initial_list)

    def __getitem__(self, i):
        """
        :rtype: ConfigVariable
        """
        return self.data[i]

    def by_name(self, name):
        """
        Finds a config variable with the matching name.

        :param str name: Variable name to find.
        :return: The matching config variable.
        :rtype: appd.model.ConfigVariable
        """
        found = [x for x in self.data if x.name == name]
        try:
            return found[0]
        except IndexError:
            raise KeyError(name)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class BusinessTransaction(JsonObject):
    """
    Represents a business transaction definition.

    .. data:: id

        Numeric ID of the BT definition.

    .. data:: name

    """
    FIELDS = {'id': '', 'name': '', 'type': 'entryPointType', 'internal_name': 'internalName',
              'is_background': 'background', 'tier_id': 'tierId', 'tier_name': 'tierName'}

    def __init__(self, bt_id=0, name='', internal_name='', tier_id=0, tier_name='',
                 bt_type='POJO', is_background=False):
        (self.id, self.name, self.internal_name, self.tier_id, self.tier_name, self.type, self.is_background) = \
            (bt_id, name, internal_name, tier_id, tier_name, bt_type, is_background)


class BusinessTransactions(JsonList):

    def __init__(self, initial_list=None):
        super(BusinessTransactions, self).__init__(BusinessTransaction, initial_list)

    def __getitem__(self, i):
        """
        :rtype: BusinessTransaction
        """
        return self.data[i]

    def by_name(self, bt_name):
        """
        Searches for business transactions that match a particular name. Note that there may be more than one
        exact match, as different tiers can have transactions with the exact same name.

        :returns: a BusinessTransactions object containing the matching business transactions.
        :rtype: BusinessTransactions
        """
        return BusinessTransactions([x for x in self.data if x.name == bt_name])

    def by_tier_and_name(self, bt_name, tier_name):
        """
        Searches for business transactions that match a particular BT name and tier name.

        :returns: a BusinessTransactions object containing the matching business transactions.
        :rtype: BusinessTransactions
        """
        return BusinessTransactions([x for x in self.data if x.name == bt_name and x.tier_name == tier_name])





"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class Application(JsonObject):
    """
    Represents a business application. The following attributes are defined:

    .. data:: id

        Numeric ID of the application.

    .. data:: name

        Application name.

    .. data:: description

        Optional description of the application.
    """

    FIELDS = {'id': '', 'name': '', 'description': ''}

    def __init__(self, app_id=0, name=None, description=None):
        self.id, self.name, self.description = app_id, name, description


class Applications(JsonList):
    """
    Represents a collection of :class:`Application` objects. Extends :class:`UserList`, so it supports the
    standard array index and :keyword:`for` semantics.
    """
    def __init__(self, initial_list=None):
        super(Applications, self).__init__(Application, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Application
        """
        return self.data[i]

    def by_name(self, name):
        """
        Finds an application by name.

        :returns: First application with the correct name
        :rtype: :class:`Application`
        """
        found = [x for x in self.data if x.name == name]
        try:
            return found[0]
        except IndexError:
            raise KeyError(name)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

from string import Template
from datetime import datetime, timedelta
from time import mktime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from MySQLdb.cursors import DictCursor
from jinja2 import Environment, FileSystemLoader
import argparse
import smtplib
import sys
import MySQLdb


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


def parse_argv():
    parser = argparse.ArgumentParser(add_help=False, usage='%(prog)s [options]',
                                     description='Generates an HTML audit report by querying the database '
                                                 'on an AppDynamics controller.')
    parser.add_argument('-h', '--host', dest='host',
                        default='localhost',
                        help='MySQL host name (default: %(default)s)')
    parser.add_argument('-P', '--port', dest='port',
                        type=int, default=3306,
                        help='MySQL server port (default: %(default)d)')
    parser.add_argument('-u', '--username', dest='username',
                        default='root',
                        help='MySQL user name (default: %(default)s)')
    parser.add_argument('-p', '--password', dest='password',
                        default='singcontroller',
                        help='MySQL password (default: %(default)s)')
    parser.add_argument('-d', '--database', dest='database',
                        default='controller',
                        help='MySQL database name (default: %(default)s)')
    parser.add_argument('-a', '--account', dest='account',
                        default='customer1',
                        help='Controller account name (default: %(default)s)')
    parser.add_argument('-n', '--days', dest='days',
                        type=int, default=1,
                        help='Number of days to report prior to today (default: %(default)d)')
    parser.add_argument('-i', '--ignore-user', dest='ignore_users',
                        action='append',
                        default=['system', 'singularity-agent', 'ldapsync'],
                        help='Users to filter from audit log (default: %(default)s)')
    parser.add_argument('-L', '--ignore-logins', dest='ignore_logins', action='store_true',
                        help='Ignore successful logins')
    parser.add_argument('-o' '--output', dest='outfile',
                        help='Send HTML output to a file instead of SMTP server')
    parser.add_argument('-f', '--from', dest='sender', default='help@appdynamics.com',
                        help='From address in email message (default: %(default)s)')
    parser.add_argument('-t', '--to', dest='to',
                        action='append',
                        default=['tradel@appdynamics.com'],
                        help='To address in email message (default: %(default)s)')
    parser.add_argument('-s', '--subject', dest='subject', default='Controller Audit Report',
                        help='Subject of email message (default: "%(default)s")')
    parser.add_argument('--smtp-server', dest='smtphost', default='localhost',
                        help='SMTP server to use (default: %(default)s)')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Enable verbose output')
    parser.add_argument('--help', action='help',
                        help='Display this help screen')
    return parser.parse_args()


def from_ts(ms):
    return datetime.fromtimestamp(ms / 1000)


def to_ts(dt):
    return int(mktime(dt.timetuple())) * 1000


ACTIONS = {
    'APP_ALERTS_CONFIGURATION': 'View Alerts',
    'APP_ANALYZE_HOME': 'Analyze',
    'APP_AUTOMATION_HOME': 'Automation',
    'APP_BACKEND_DASHBOARD|OVERVIEW_TAB': 'Backend Dashboard',
    'APP_BACKEND_LIST': 'Backends List',
    'APP_BASELINES': 'Baselines',
    'APP_BT_DETAIL': 'BT Dashboard',
    'APP_BT_DETAIL|OVERVIEW_TAB': 'BT Dashboard',
    'APP_BT_LIST': 'BT List',
    'APP_BT_LIST|OVERVIEW_TAB': 'BT List',
    'APP_COMPARE_METRICS': 'Correlation Analysis',
    'APP_COMPONENT_MANAGER': 'Tier Dashboard',
    'APP_COMPONENT_MANAGER|OVERVIEW_TAB': 'Tier Dashboard',
    'APP_CONFIGURATION': 'Configure',
    'APP_CONFIGURATION_ACTIONS': 'Actions',
    'APP_CONFIGURATION_HOME': 'Configure',
    'APP_CPU_VS_LOAD': 'Scalability Analysis',
    'APP_DASHBOARD': 'Application Dashboard',
    'APP_DASHBOARD|OVERVIEW_TAB': 'Application Dashboard',
    'APP_DATABASE_LIST': 'Databases',
    'APP_ENVIRONMENT_PROPERTIES'
    'APP_ERRORS': 'Errors',
    'APP_ERRORS|OVERVIEW_TAB': 'Errors',
    'APP_EVENTSTREAM_LIST': 'Events',
    'APP_IIS_APP_POOLS': 'IIS App Pools',
    'APP_INCIDENT_LIST': 'Policy Violations',
    'APP_INFO_POINT_LIST': 'Information Points',
    'APP_INFRA_MESSAGE_SERVERS': 'Message Servers',
    'APP_INFRA_OTHER_SERVERS': 'Other Backends',
    'APP_INFRASTRUCTURE': 'Servers',
    'APP_NODE_MANAGER': 'Node Dashboard',
    'APP_NODE_MANAGER|OVERVIEW_TAB': 'Node Dashboard',
    'APP_POLICY_LIST': 'Policies',
    'APP_RELEASE_ANALYSIS': 'Compare Releases',
    'APP_REPORTING': 'Reports',
    'APP_REQUEST_LIST': 'Requests',
    'APP_RESPONSE_TIME_VS_LOAD': 'Scalability Analysis',
    'APP_SCHEDULE_LIST': 'Schedules',
    'APP_SERVER_HOME': 'Home',
    'APP_SLOW_RESPONSE_TIMES': 'Slow Response times',
    'APP_TASK_LIST': 'Tasks',
    'APP_THRESHOLD_LIST': 'Thresholds',
    'APP_TROUBLESHOOT_HOME': 'Troubleshooting',
    'APP_WORKFLOW_EXECUTION_LIST': 'Workflow Executions',
    'APP_WORKFLOW_LIST': 'Workflows',
    'FLOW_ICON_MOVED': 'Flow Map Changed',
    'LOGIN': 'Login Success',
    'LOGIN_FAILED': 'Login Failure',
    'OBJECT_CREATED': 'Object Created',
    'OBJECT_DELETED': 'Object Deleted',
    'OBJECT_UPDATED': 'Object Changed',
    'USER_PASSWORD_CHANGED': 'Password Changed'
}

AUDIT_TABLES = {
    'AGENT_CONFIGURATION': {'select_expr': "CONCAT(t.agent_type, ' ', t.entity_type)",
                            'display_name': 'Agent Configuration'},
    'APPLICATION': {'display_name': 'Application',
                    'app_name_expr': 'name',
                    'app_id_expr': 'id'},
    'APPLICATION_COMPONENT_NODE': {'display_name': 'Node',
                                   'app_name_expr': 'NULL',
                                   'app_id_expr': 'NULL'},
    'APPLICATION_COMPONENT': {'display_name': 'Tier'},
    'APPLICATION_DIAGNOSTIC_DATA': {},
    'BACKEND_DISCOVERY_CONFIG': {'display_name': 'Backend Detection Config'},
    'BUSINESS_TRANSACTION': {'join_to': 'abusiness_transaction',
                             'display_name': 'Business Transaction'},
    'BUSINESS_TRANSACTION_GROUP': {'join_to': 'abusiness_transaction',
                                   'display_name': 'Business Transaction Group'},
    'CALL_GRAPH_CONFIGURATION': {'select_expr': "'CALL_GRAPH_CONFIGURATION'",
                                 'display_name': 'Call Graph Config'},
    'CUSTOM_EXIT_POINT_DEFINITION': {'display_name': 'Custom Backend Config'},
    'CUSTOM_MATCH_POINT_DEFINITION': {'select_expr': "CONCAT(t.name, ' [', t.entry_point_type, ']')",
                                      'display_name': 'Custom Entry Point'},
    'DASHBOARD': {'display_name': 'Custom Dashboard'},
    'DOT_NET_ERROR_CONFIGURATION': {'select_expr': "'DOT_NET_ERROR_CONFIGURATION'",
                                    'join_to': 'dotnet_error_configuration',
                                    'display_name': 'Error Detection Config [.NET]'},
    'ERROR_CONFIGURATION': {'select_expr': "'ERROR_CONFIGURATION'",
                            'display_name': 'Error Detection Config'},
    'EUM_CONFIGURATION': {'select_expr': "'EUM_CONFIGURATION'",
                          'display_name': 'EUM Config'},
    # 'GLOBAL_CONFIGURATION': ('name', 'GLOBAL_CONFIGURATION'),
    'HTTP_REQUEST_DATA_GATHERER_CONFIG': {'join_to': 'adata_gatherer_config',
                                          'display_name': 'HTTP Data Collector Config'},
    'JMX_CONFIG': {'join_to': 'jmx_rule'},
    'MEMORY_CONFIGURATION': {'select_expr': "'MEMORY_CONFIGURATION'",
                             'display_name': 'Memory Monitoring Config'},
    'POJO_DATA_GATHERER_CONFIG': {'join_to': 'adata_gatherer_config',
                                  'display_name': 'Custom Method Collector Config'},
    'POLICY': {'display_name': 'Policy'},
    # 'RULE': ('name', 'RULE'),
    'SQL_DATA_GATHERER_CONFIG': {'join_to': 'adata_gatherer_config',
                                 'display_name': 'SQL Data Collector Config'},
    'TRANSACTION_MATCH_POINT_CONFIG': {'display_name': 'Transaction Detection Config'},
    'USER': {'display_name': 'User'},
}

ENTITY_TYPES = {'APPLICATION': {'display_name': 'Application',
                                'app_name_expr': 't.name',
                                'app_id_expr': 't.id'},
                'APPLICATION_COMPONENT': {'display_name': 'Tier'},
                'APPLICATION_COMPONENT_NODE': {'display_name': 'Node'}}


def connect(args):
    conn = MySQLdb.connect(args.host, args.username, args.password, args.database, args.port,
                           cursorclass=DictCursor)
    cur = conn.cursor()
    return conn, cur


def create_temp_table(cur):
    try:
        sql = """
        CREATE TEMPORARY TABLE audit_report (
            ts_ms bigint,
            account_name varchar(100),
            account_id int,
            user_name varchar(100),
            user_security_provider_type varchar(25),
            user_id int,
            object_name varchar(100),
            object_id int,
            application_name varchar(100),
            application_id int,
            action varchar(100),
            execution_time_ms bigint,
            object_desc varchar(1000),
            INDEX ts_ms (ts_ms)) ENGINE=memory DEFAULT CHARSET=utf8;
        """
        cur.execute(sql)
    except:
        print("*** ERROR creating temporary table", file=sys.stderr)
        raise


def insert_object_crud(cur, params):
    try:
        for table_name, table_data in AUDIT_TABLES.items():
            params.update({
                'table_name': table_name,
                'id_field': table_data.get('id_field', 'id'),
                'select_expr': table_data.get('select_expr', 'name'),
                'app_name_expr': table_data.get('app_name_expr', 'NULL'),
                'app_id_expr': table_data.get('app_id_expr', 'NULL'),
                'join_to': table_data.get('join_to', table_name),
            })

            sql = """
            INSERT INTO audit_report
            (SELECT ca.ts_ms, ca.account_name, ca.account_id, ca.user_name, ca.user_security_provider_type,
                ca.user_id, ca.object_name, ca.object_id, $app_name_expr, $app_id_expr,
                ca.action, ca.execution_time_ms,
                $select_expr
            FROM controller_audit ca
            JOIN $join_to t ON ($id_field = ca.object_id)
            WHERE LOWER(ca.account_name) = LOWER('$acct')
              AND LOWER(ca.object_name) = '$table_name'
              AND ca.action LIKE 'OBJECT%'
              AND ca.object_name NOT IN ('AGENT_CONFIGURATION')
              AND ca.user_name NOT IN ($users_to_ignore)
              AND ca.ts_ms BETWEEN $start_ts AND $end_ts
            ORDER BY ca.ts_ms)
        """
            sql = Template(sql).substitute(params)
            cur.execute(sql)
    except:
        print("*** ERROR EXECUTING SQL: ", sql, file=sys.stderr)
        raise


def insert_agent_configuration_crud(cur, params):
    try:
        for entity_type, entity_data in ENTITY_TYPES.items():
            params.update({
                'entity_type': entity_type,
                'app_name_expr': entity_data.get('app_name_expr', 'NULL'),
                'app_id_expr': entity_data.get('app_id_expr', 'NULL'),
                'display_name': entity_data.get('display_name', entity_type)
            })
            sql = """
        INSERT INTO audit_report
        (SELECT ca.ts_ms, ca.account_name, ca.account_id, ca.user_name, ca.user_security_provider_type,
               ca.user_id, ca.object_name, ca.object_id, $app_name_expr, $app_id_expr, ca.action, ca.execution_time_ms,
               CONCAT('$display_name', ' ', t.name) AS display_name
                FROM controller_audit ca
                JOIN agent_configuration ac ON (ac.id = ca.object_id)
                JOIN $entity_type t ON (t.id = ac.entity_id)
                WHERE LOWER(ca.account_name) = LOWER('$acct')
                  AND LOWER(ca.object_name) = 'AGENT_CONFIGURATION'
                  AND ca.action LIKE 'OBJECT%%'
                  AND ca.user_name NOT IN ($users_to_ignore)
                  AND ca.ts_ms BETWEEN $start_ts AND $end_ts
                ORDER BY ca.ts_ms)
        """
            sql = Template(sql).substitute(params)
            cur.execute(sql)
    except:
        print("*** ERROR EXECUTING SQL: ", sql, file=sys.stderr)
        raise


def insert_logins(cur, params):
    try:
        sql = """
        INSERT INTO audit_report
        (SELECT ca.ts_ms, ca.account_name, ca.account_id, ca.user_name, ca.user_security_provider_type,
           ca.user_id, ca.object_name, ca.object_id, NULL, NULL, ca.action, ca.execution_time_ms,
           NULL AS display_name
            FROM controller_audit ca
            WHERE LOWER(ca.account_name) = LOWER('$acct')
              AND ca.action IN ($login_types)
              AND ca.user_name NOT IN ($users_to_ignore)
              AND ca.ts_ms BETWEEN $start_ts AND $end_ts
            ORDER BY ca.ts_ms)
    """
        sql = Template(sql).substitute(params)
        cur.execute(sql)
    except:
        print("*** ERROR EXECUTING SQL: ", sql, file=sys.stderr)
        raise


def select_results(cur):
    sql = """
    SELECT ts_ms, account_name, account_id, user_name, user_security_provider_type,
        user_id, object_name, object_id, application_id, action, execution_time_ms, object_desc
    FROM audit_report
    ORDER BY ts_ms
    """
    cur.execute(sql)


def generate_output(args, params):
    env = Environment(loader=FileSystemLoader(['templates', 'appd/reports/templates', '.']))
    template = env.get_template('audit.jinja.html')
    return template.render(params)


def populate_params(args):
    end_time = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    start_time = end_time + timedelta(-args.days)
    login_types = ['LOGIN_FAILED']
    if not args.ignore_logins:
        login_types.append('LOGIN')
    return {
        'end_time': end_time,
        'start_time': start_time,
        'start_ts': to_ts(start_time),
        'end_ts': to_ts(end_time),
        'acct': args.account,
        'users_to_ignore': ', '.join(["'" + x + "'" for x in args.ignore_users]),
        'login_types': ', '.join(["'" + x + "'" for x in login_types])
    }


def send_html_email(args, html):
    msg = MIMEMultipart('alternative')
    msg['Subject'] = args.subject
    msg['From'] = args.sender
    msg['To'] = ', '.join(args.to)

    part1 = MIMEText('Please use an HTML email client to view this message.', 'plain')
    part2 = MIMEText(html, 'html', 'utf-8')
    msg.attach(part1)
    msg.attach(part2)

    s = smtplib.SMTP(args.smtphost)
    if args.verbose:
        s.set_debuglevel(1)
    s.sendmail(args.sender, args.to, msg.as_string())
    s.close()


if __name__ == '__main__':

    args = parse_argv()
    params = populate_params(args)

    (conn, cur) = connect(args)

    create_temp_table(cur)
    insert_object_crud(cur, params)
    insert_agent_configuration_crud(cur, params)
    insert_logins(cur, params)
    select_results(cur)

    start_time = params['start_time']
    end_time = params['end_time']
    html = generate_output(args, locals())
    conn.close()

    if args.outfile:
        with open(args.outfile, 'w') as f:
            f.write(html)
    else:
        send_html_email(args, html)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class MetricTreeNode(JsonObject):

    FIELDS = {'name': '', 'type': ''}
    NODE_TYPES = ('leaf', 'folder')

    def __init__(self, parent=None, node_name='', node_type=''):
        self.parent, self.name, self.type = parent, node_name, node_type
        self._children = MetricTreeNodes()
        if parent:
            parent._children.append(self)

    @property
    def path(self):
        n = self
        stack = []
        while n:
            stack.insert(0, n.name)
            n = n.parent
        return '|'.join(stack)


class MetricTreeNodes(JsonList):

    def __init__(self, initial_list=None, parent=None):
        super(MetricTreeNodes, self).__init__(MetricTreeNode, initial_list)

        self.parent = parent
        if parent:
            if not isinstance(parent, MetricTreeNode):
                raise TypeError('was expecting a MetricTreeNode')
            for x in self.data:
                x.parent = parent

    def __getitem__(self, i):
        """
        :rtype: MetricTreeNode
        """
        return self.data[i]

    @classmethod
    def from_json(cls, json_list, parent=None):
        return cls(json_list, parent)

    def by_name(self, name):
        """
        Finds a tree node with the matching name.

        :param str name: Variable name to find.
        :return: Metric tree node matching the name.
        :rtype: appd.model.MetricTreeNode
        """
        found = [x for x in self.data if x.name == name]
        try:
            return found[0]
        except IndexError:
            raise KeyError(name)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class EntityDefinition(JsonObject):
    FIELDS = {'entity_id': 'entityId',
              'type': 'entityType'}

    ENTITY_TYPES = ('ACCOUNT', 'AGENT_CONFIGURATION', 'ALL', 'APPLICATION', 'APPLICATION_COMPONENT',
                    'APPLICATION_COMPONENT_NODE', 'APPLICATION_DIAGNOSTIC_DATA', 'AUTOMATIC_LEAK_DETECTION',
                    'BACKEND', 'BACKEND_DISCOVERY_CONFIG', 'BUSINESS_TRANSACTION', 'BUSINESS_TRANSACTION_GROUP',
                    'CALL_GRAPH_CONFIGURATION', 'CUSTOM_EXIT_POINT_DEFINITION', 'CUSTOM_MATCH_POINT_DEFINITION',
                    'CUSTOM_MEMORY_STRUCTURE', 'DASHBOARD', 'ERROR', 'ERROR_CONFIGURATION',
                    'DOT_NET_ERROR_CONFIGURATION', 'PHP_ERROR_CONFIGURATION', 'EUM_CONFIGURATION', 'EVENT',
                    'GLOBAL_CONFIGURATION', 'INCIDENT', 'JMX_CONFIG', 'MACHINE_INSTANCE', 'MEMORY_CONFIGURATION',
                    'NOTIFICATION_CONFIG', 'OBJECT_INSTANCE_TRACKING', 'POLICY', 'RULE', 'SQL_DATA_GATHERER_CONFIG',
                    'STACK_TRACE', 'THREAD_TASK', 'TRANSACTION_MATCH_POINT_CONFIG', 'USER', 'GROUP', 'ACCOUNT_ROLE',
                    'WORKFLOW', 'WORKFLOW_EXCUTION', 'POJO_DATA_GATHERER_CONFIG', 'HTTP_REQUEST_DATA_GATHERER_CONFIG',
                    'BASE_PAGE', 'IFRAME', 'AJAX_REQUEST', 'INFO_POINT')

    def __init__(self, entity_id=0, entity_type=''):
        self._type = None
        (self.id, self.type) = (entity_id, entity_type)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, new_type):
        self._list_setter('_type', new_type, EntityDefinition.ENTITY_TYPES)



#!/bin/bash

from appd.request import AppDynamicsClient
from appd.time import from_ts, to_ts
from datetime import datetime
from pprint import pprint

appdProduction = AppDynamicsClient('https://fedex1.saas.appdynamics.com', 'RetailRelyAPI', 'F3d#x@M!45', 'fedex1' )

for app in appdProduction.get_applications():
    if 'OFFICE' in app.name:
        appid = app.id
        print(app.name, app.id)

officeNodes = appdProduction.get_nodes(app_id=appid)
hrviolations = appdProduction.get_healthrule_violations(app_id=appid,time_range_type='BEFORE_NOW',duration_in_mins=15)
pprint(hrviolations)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sample script to print a list of events from the specified controller and app.
"""

from __future__ import print_function

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient

__author__ = 'Kyle Furlong'
__copyright__ = 'Copyright (c) 2013-2017 AppDynamics Inc.'


args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)

resp = c.create_event(app_id=16,
                      summary='Event 1',
                      comment='This is an event created by the Python SDK',
                      severity='INFO',
                      eventtype='APPLICATION_CONFIG_CHANGE')
print(resp)

resp = c.create_event(app_id=16,
                      summary='Custom Event 1',
                      comment='This is an event created by the Python SDK',
                      severity='INFO',
                      eventtype='CUSTOM',
                      customeventtype='MYCUSTOMEVENT',
                      node="python-node-1",
                      tier="python",
                      bt="/admin")
print(resp)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class Account(JsonObject):
    """
    Represents a tenant account on the controller. The following attributes are defined:

    .. data:: id

        Numeric ID of the account.

    .. data:: name

        Account name.
    """

    FIELDS = {'id': '', 'name': ''}

    def __init__(self, acct_id='0', name=None):
        self.id, self.name = acct_id, name


class Accounts(JsonList):
    """
    Represents a collection of :class:Account objects. Extends :class:UserList, so it supports the
    standard array index and :keyword:`for` semantics.
    """
    def __init__(self, initial_list=None):
        super(Accounts, self).__init__(Account, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Account
        """
        return self.data[i]

    def by_name(self, name):
        """
        Finds an account by name.

        :returns: First account with the correct name
        :rtype: Account
        """
        found = [x for x in self.data if x.name == name]
        try:
            return found[0]
        except IndexError:
            raise KeyError(name)



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Kyle Furlong <kyle.furlong@appdynamics.com>
"""

from . import JsonObject, JsonList


class ActionSuppression(JsonObject):

    FIELDS = {'id': '',
              'timeRange': '',
              'name': '',
              'affects': ''}

    def __init__(self, id=0, timeRange=None, name='', affects=None):
        (self.id, self.timeRange, self.name, self.affects) = (id, timeRange, name, affects)


class ActionSuppressions(JsonList):

    def __init__(self, initial_list=None):
        super(ActionSuppressions, self).__init__(ActionSuppression, initial_list)

    def __getitem__(self, i):
        """
        :rtype: ActionSuppression
        """
        return self.data[i]


class ActionSuppressionsResponse(JsonObject):

    FIELDS = {}

    def __init__(self):
        self.actionSuppressions = ActionSuppressions()

    @classmethod
    def from_json(cls, json_dict):
        print(json_dict)
        obj = super(ActionSuppressionsResponse, cls).from_json(json_dict)
        if 'actionSuppressions' in json_dict:
            obj.actionSuppressions = ActionSuppressions.from_json(json_dict['actionSuppressions'])
        return obj



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Kyle Furlong <kyle.furlong@appdynamics.com>
"""

from . import JsonObject, JsonList


class SetControllerUrl(JsonObject):

    FIELDS = {'controllerURL': ''}

    def __init__(self, controllerURL=''):
        (self.controllerURL) = (controllerURL)


class SetControllerUrlResponse(JsonList):

    def __init__(self, initial_list=None):
        super(SetControllerUrlResponse, self).__init__(SetControllerUrl, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Node
        """
        return self.data[i]



#! /usr/bin/env python2.7
# -*- coding: utf-8 -*-

# __author__ = 'Todd Radel <tradel@appdynamics.com>'

from setuptools import setup
import io
import os

import appd

here = os.path.abspath(os.path.dirname(__file__))


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


setup(name='AppDynamicsRESTx',
      version=appd.__version__,
      description='AppDynamics REST API Library',
      long_description=read('README.rst'),
      author='Todd Radel',
      author_email='tradel@appdynamics.com',
      maintainer='Mike Phillipson',
      maintainer_email='MICHAEL_PHILLIPSON1@homedepot.com',
      url='https://github.com/homedepot/AppDynamicsRESTx',
      packages=['appd','appd.model'],
      platforms='any',
      package_data={'': ['README.rst', 'data/*', 'examples/*', 'templates/*']},
      install_requires=['requests', 'argparse', 'future', 'pycurl'],
      extras_require={'examples': ['lxml', 'tzlocal', 'jinja2'], 'testing': ['nose']},
      test_suite='nose.collector',
      tests_require=['nose'],
      license='Apache',
      classifiers=[
          'Programming Language :: Python',
          'Intended Audience :: Developers',
          'Development Status :: 4 - Beta',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: System :: Monitoring',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8'
      ])



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList


class Tier(JsonObject):

    FIELDS = {'id': '', 'name': '', 'description': '', 'type': '',
              'node_count': 'numberOfNodes', 'agent_type': 'agentType'}
    AGENT_TYPES = ('APP_AGENT', 'MACHINE_AGENT',
                   'DOT_NET_APP_AGENT', 'DOT_NET_MACHINE_AGENT',
                   'PHP_APP_AGENT', 'PHP_MACHINE_AGENT',
                   'NODEJS_APP_AGENT','NODEJS_MACHINE_AGENT',
                   'PYTHON_APP_AGENT',
                   'NATIVE_APP_AGENT','NATIVE_SDK', 'NATIVE_DYNAMIC','NATIVE_WEB_SERVER',
                   'DB_AGENT','DB_COLLECTOR',
                   'RUBY_APP_AGENT',
                   'SIM_MACHINE_AGENT','APM_MACHINE_AGENT','SERVICE_AVAIL_MACHINE_AGENT',
                   'APM_APP_AGENT',
                   'ANALYTICS_AGENT',
                   'GOLANG_SDK',
                   'WMB_AGENT')

    def __init(self, tier_id=0, name='', description='', agent_type='JAVA_AGENT', node_count=0,
               tier_type='Java Application Server'):
        self._agent_type = None
        self.id, self.name, self.description, self.agent_type, self.node_count, self.type = \
            tier_id, name, description, agent_type, node_count, tier_type

    @property
    def agent_type(self):
        """
        :rtype: str
        """
        return self._agent_type

    @agent_type.setter
    def agent_type(self, agent_type):
        self._list_setter('_agent_type', agent_type, Tier.AGENT_TYPES)


class Tiers(JsonList):

    def __init__(self, initial_list=None):
        super(Tiers, self).__init__(Tier, initial_list)

    def __getitem__(self, i):
        """
        :rtype: Tier
        """
        return self.data[i]

    def by_agent_type(self, agent_type):
        """
        Searches for tiers of a particular type (which should be one of Tier.AGENT_TYPES). For example, to find
        all the Java app server tiers:

        >>> from appd.request import AppDynamicsClient
        >>> client = AppDynamicsClient(...)
        >>> all_tiers = client.get_tiers()
        >>> java_tiers = all_tiers.by_agent_type('APP_AGENT')

        :returns: a Tiers object containing any tiers matching the criteria
        :rtype: Tiers
        """
        return Tiers([x for x in self.data if x.agentType == agent_type])



"""
Model classes for AppDynamics REST API

.. moduleauthor:: Todd Radel <tradel@appdynamics.com>
"""

from . import JsonObject, JsonList
from .metric_value import MetricValues

class MetricDataSingle(JsonObject):

    FIELDS = {
        'frequency': '',
        'path': 'metricPath'
    }

    FREQUENCIES = ('ONE_MIN', 'TEN_MIN', 'SIXTY_MIN')

    def __init__(self, path='', frequency='ONE_MIN', values=MetricValues()):
        self._frequency = None
        self.path, self.frequency, self.values = path, frequency, values

    @classmethod
    def _set_fields_from_json_dict(cls, obj, json_dict):
        JsonObject._set_fields_from_json_dict(obj, json_dict)
        obj.values = MetricValues.from_json(json_dict['metricValues'])

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, new_freq):
        self._list_setter('_frequency', new_freq, MetricDataSingle.FREQUENCIES)


class MetricData(JsonList):
    def __init__(self, initial_list=None):
        super(MetricData, self).__init__(MetricDataSingle, initial_list)

    def __getitem__(self, i):
        """
        :rtype: MetricDataSingle
        """
        return self.data[i]

    def by_partial_name(self, name):
        return MetricData([x for x in self if name in x.path])

    def by_leaf_name(self, name):
        return MetricData([x for x in self if x.path.split('|')[-1] == name])

    def by_path(self, path):
        return MetricData([x for x in self if x.path == path])

    def first_value(self):
        return self[0].values[0].value



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This script generates a report of response times, calls, and errors for all the backends registered
on a controller. The output is generated as XML and can be sent to a file with this syntax::

    python BackendMetrics.py > metric_output.xml

By default this will connect to a controller on localhost:8090. The script accepts command-line
arguments to change the connection parameters:

--controller=<url>
--account=<account>
--username=<name>
--password=<password>

Example::
    python BackendMetrics.py --controller=http://10.1.2.3:8090 --account=customer1 --username=demo --password=abc123
"""

from __future__ import print_function

from collections import defaultdict
from datetime import datetime
from time import mktime
from lxml.builder import ElementMaker
from lxml import etree
import tzlocal

from appd.cmdline import parse_argv
from appd.request import AppDynamicsClient


__author__ = 'Todd Radel'
__copyright__ = 'Copyright (c) 2013-2015 AppDynamics Inc.'


# The report will generate data for the 24-hour period before midnight of the current day. To change the
# reporting period, adjust these variables.

time_in_mins = 24 * 60
end_time = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
end_epoch = int(mktime(end_time.timetuple())) * 1000


# Metric mapper
#
# Uses a hash table to call the function appropriate for specific metric types, e.g.: calls map_art for
# Average Response Time, map_cpm for Calls per Minute, etc.

def map_art(d, metric_total, metric_average, metric_sum, max_point):
    d['art'] = metric_average


def map_cpm(d, metric_total, metric_average, metric_sum, max_point):
    d['total_calls'] = metric_total
    d['cpm_average'] = metric_average
    d['cpm_max'] = max_point[0]
    d['cpm_max_time'] = max_point[1]


def map_err(d, metric_total, metric_average, metric_sum, max_point):
    d['total_errors'] = metric_total
    d['epm_avg'] = metric_average
    d['epm_max'] = max_point[0]
    d['epm_max_time'] = max_point[1]


def no_such_metric(d, metric_total, metric_average, metric_sum, max_point):
    raise ValueError("no such metric")


METRIC_DISPATCHER = {
    'Average Response Time (ms)': map_art,
    'Calls per Minute': map_cpm,
    'Errors per Minute': map_err,
    'DEFAULT': no_such_metric
}


# Helper functions

def now_rfc3339():
    return datetime.now(tzlocal.get_localzone()).isoformat('T')


def freq_to_mins(md):
    FREQ_MAP = {'ONE_MIN': 1, 'TEN_MIN': 10, 'SIXTY_MIN': 60}
    return FREQ_MAP[md.frequency]


# Parse command line arguments and create AD client:

args = parse_argv()
c = AppDynamicsClient(args.url, args.username, args.password, args.account, args.verbose)


# Get the list of configured apps, and get backend metrics for each one:

rows = defaultdict(dict)
for app in c.get_applications():
    for md in c.get_metrics('Backends|*|*', app.id, time_range_type='BEFORE_TIME', end_time=end_epoch,
                            duration_in_mins=time_in_mins, rollup=False):

        # Get the last two components of the metric path. This should be 'backend_name|metric_name'.
        backend_name, metric_name = md.path.split('|')[-2:]

        if 'Discovered backend call' in backend_name:
            backend_name = backend_name[26:]
            metric_sum = sum([x.value for x in md.values])
            metric_total = metric_sum * freq_to_mins(md)
            metric_average = 0
            max_point = (0, None)
            if len(md.values) > 0:
                metric_average = metric_sum / len(md.values)
                max_point = max([(max(x.value, x.current, x.max), x.start_time) for x in md.values])

            func = METRIC_DISPATCHER.get(metric_name, None)
            if func:
                func(rows[backend_name], metric_total, metric_average, metric_sum, max_point)
                rows[backend_name]['app'] = app.name

# Generate the report.

XSI = 'http://www.w3.org/2001/XMLSchema-instance'
E = ElementMaker(nsmap={'xsi': XSI})

root = E.BackendResponseTimes(Controller=c.base_url, GenerationTime=now_rfc3339())
root.set('{%s}noNamespaceSchemaLocation' % XSI, 'backend_metrics.xsd')

for k, v in sorted(rows.items()):
    v.setdefault('cpm_max_time', '')
    v.setdefault('epm_max_time', '')
    root.append(E.Backend(
        E.ApplicationName(v['app']),
        E.BackendName(k),
        E.AverageResponseTime(str(v.get('art', 0))),
        E.CallsPerMinute(str(v.get('cpm_average', 0))),
        E.TotalCalls(str(v.get('total_calls', 0))),
        E.MaximumCallsPerMinute(str(v.get('cpm_max', 0))),
        E.MaximumCallTime(v['cpm_max_time'].isoformat('T') if v['cpm_max_time'] else ''),
        E.ErrorsPerMinute(str(v.get('epm_avg', 0))),
        E.TotalErrors(str(v.get('total_errors', 0))),
        E.MaximumErrorsPerMinute(str(v.get('epm_max', 0))),
        E.MaximumErrorTime(v['epm_max_time'].isoformat('T') if v['epm_max_time'] else ''),
    ))

# Print the report to stdout.

print(etree.ProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"'))
print(etree.tostring(root, pretty_print=True, encoding='UTF-8'))
